"""
Tests for CALIB module

FIXME need satrecov tests
FIXME need sensitivity correction tests
FIXME need test with edge correction
"""
import math
from six import StringIO

import pytest
import numpy as np

from fsl.data.image import Image

from oxasl import Workspace, calib, AslImage

def _get_imgs(shape=(5, 5, 5)):
    """
    Create test images
    """
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    return perf_img, calib_img

def _expected_m0(ref_mean, t1, t2, pc, gain=1.0, alpha=1.0, te=0, tr=3.2, t2b=150):
    """
    Expected M0 value from reference region mean
    """
    #print("ref mean: ", ref_mean)
    m0 =  ref_mean / (1 - math.exp(- tr / t1))
    #print("T1 correction: ", m0)
    m0 = m0 / math.exp(-te/t2)
    m0 = m0 * gain / pc
    m0 = m0 * math.exp(-te/t2b)
    #print("T2 correction: ", m0)
    m0 *= alpha
    #print("IE correction: ", m0)
    return m0

def test_nodata():
    """
    Check we get an error if there is no perfusion data
    """
    d = np.random.rand(5, 5, 5, 6)
    calib_img = Image(name="calib", image=d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    with pytest.raises(ValueError):
        calib.calibrate(wsp, None)

def test_nocalib():
    """
    Check we get an error if there is no calibration data
    """
    d = np.random.rand(5, 5, 5, 6)
    perf_img = Image(name="perfusion", image=d)

    wsp = Workspace(calib_method="voxelwise")
    with pytest.raises(ValueError):
        calib.calibrate(wsp, perf_img)

def test_bad_method():
    """
    Check we get an error on an invalid method
    """
    d = np.random.rand(5, 5, 5, 6)
    perf_img = Image(name="perfusion", image=d)

    d = np.random.rand(5, 5, 5, 6)
    calib_img = Image(name="calib", image=d)

    wsp = Workspace(calib=calib_img, calib_method="random")
    with pytest.raises(ValueError):
        calib.calibrate(wsp, perf_img)

def test_defaults():
    """
    Check default calibration works
    """
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * perf_d / calib_d)

def test_cgain():
    """
    Check calibration gain
    """
    GAIN = 1.23
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", calib_gain=GAIN)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 / GAIN * perf_d / calib_d)

def test_alpha():
    """
    Check inversion efficiency
    """
    ALPHA = 0.74
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    perf_calib = calib.calibrate(wsp, perf_img, alpha=ALPHA)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 / ALPHA * perf_d / calib_d)

def test_multiplier():
    """
    Check end multiplier for physical units
    """
    MULTIPLIER = 7654
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    perf_calib = calib.calibrate(wsp, perf_img, multiplier=MULTIPLIER)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * MULTIPLIER * perf_d / calib_d)

def test_partition_coeff():
    """
    Check tissue/blood partition coefficient
    """
    PC = 0.67
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", pct=PC)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, PC * perf_d / calib_d)

def test_shorttr_corr():
    """
    Check correction for short TR
    """
    TR, T1 = 3.0, 1.1
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", tr=TR, t1t=T1)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    factor =  1 - math.exp(-TR / T1)
    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * factor * perf_d / calib_d)

def test_shorttr_corr_not1():
    """
    Check correction for short TR is not done (and generates warning) if T1 not given
    """
    TR = 3.0
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    log = StringIO()
    wsp = Workspace(calib=calib_img, calib_method="voxelwise", tr=TR, log=log)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)
    assert("WARNING" in log.getvalue())

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * perf_d / calib_d)

def test_defaults_var():
    """
    Check default calibration works on variances
    """
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    perf_calib = calib.calibrate(wsp, perf_img, var=True)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * 0.9 * perf_d / np.square(calib_d))

def test_cgain_var():
    """
    Check calibration gain on variances
    """
    GAIN = 1.23
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", calib_gain=GAIN)
    perf_calib = calib.calibrate(wsp, perf_img, var=True)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * 0.9 / GAIN / GAIN * perf_d / calib_d / calib_d)

def test_alpha_var():
    """
    Check inversion efficiency on variances
    """
    ALPHA = 0.74
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise",)
    perf_calib = calib.calibrate(wsp, perf_img, var=True, alpha=ALPHA)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * 0.9 / ALPHA /ALPHA * perf_d / calib_d / calib_d)

def test_multiplier_var():
    """
    Check end multiplier for physical units on variances
    """
    MULTIPLIER = 7654
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise")
    perf_calib = calib.calibrate(wsp, perf_img, var=True, multiplier=MULTIPLIER)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * 0.9 * MULTIPLIER * MULTIPLIER * perf_d / calib_d / calib_d)

def test_partition_coeff_var():
    """
    Check tissue/blood partition coefficient on variances
    """
    PC = 0.67
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", pct=PC)
    perf_calib = calib.calibrate(wsp, perf_img, var=True)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, PC * PC * perf_d / calib_d / calib_d)

def test_shorttr_corr_var():
    """
    Check correction for short TR on variances
    """
    TR, T1 = 3.0, 1.1
    perf_d = np.random.rand(5, 5, 5)
    perf_img = Image(name="perfusion", image=perf_d)

    calib_d = np.random.rand(5, 5, 5)
    calib_img = Image(name="calib", image=calib_d)

    wsp = Workspace(calib=calib_img, calib_method="voxelwise", tr=TR, t1t=T1)
    perf_calib = calib.calibrate(wsp, perf_img, var=True)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    factor =  1 - math.exp(-TR / T1)
    # Default partition coefficient is 0.9
    np.testing.assert_allclose(calibrated_d, 0.9 * 0.9 * factor * factor * perf_d / calib_d / calib_d)

def test_refregion_defaults():
    """
    Reference region defaults to CSF
    """
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data
    assert(perf_calib.name == "perfusion_calib")
    assert(perf_calib.shape == perf_img.shape)

    # CSF defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 4.3, 750, 1.15)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_unknown():
    """
    Reference region unknown type
    """
    perf_img, calib_img = _get_imgs()
    
    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="kryptonite")
    with pytest.raises(ValueError):
        calib.calibrate(wsp, perf_img)
    
def test_refregion_csf_all():
    """
    Reference region mask is whole image
    """
    perf_img, calib_img = _get_imgs()
    
    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="csf")
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # CSF defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 4.3, 750, 1.15)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_wm_all():
    """
    Reference region mask is whole image
    """
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm")
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # WM defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_gm_all():
    """
    Reference region mask is whole image
    """
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="gm")
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # GM defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.3, 100, 0.98)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_tr():
    """
    Reference region override TR
    """
    TR = 7.1
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", tr=TR)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # WM defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82, tr=TR)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_te():
    """
    Reference region override TE
    """
    TE = 1.2
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", te=TE)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82, te=TE)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_t2b():
    """
    Reference region override T2b
    """
    T2b = 782
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", t2b=T2b)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82, t2b=T2b)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_t1r():
    """
    Reference region override T1 of ref tissue
    """
    T1 = 1.03
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", t1r=T1)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), T1, 50, 0.82)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_t2r():
    """
    Reference region override T2 of ref tissue
    """
    T2 = 635
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", t2r=T2)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, T2, 0.82)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_pc():
    """
    Reference region override partition coeff of ref tissue
    """
    PC = 0.67
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", pcr=PC)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, PC)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_csf_t2star():
    """
    Using T2* instead of T2 for CSF
    """
    perf_img, calib_img = _get_imgs()
    
    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="csf", t2star=True)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 4.3, 400, 1.15)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_wm_t2star():
    """
    Using T2* instead of T2 for WM
    """
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", t2star=True)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # WM defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_gm_t2star():
    """
    Using T2* instead of T2 for GM
    """
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="gm", t2star=True)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    # GM defaults
    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.3, 60, 0.98)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_gain():
    """
    Modify calibration gain
    """
    GAIN = 3.4
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm", calib_gain=GAIN)
    perf_calib = calib.calibrate(wsp, perf_img)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82, gain=GAIN)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)

def test_refregion_alpha():
    """
    Modify calibration gain
    """
    ALPHA = 0.46
    perf_img, calib_img = _get_imgs()

    ref_d = np.ones((5, 5, 5))
    ref_img = Image(name="refmask", image=ref_d)

    wsp = Workspace(calib=calib_img, calib_method="refregion", refmask=ref_img, tissref="wm")
    perf_calib = calib.calibrate(wsp, perf_img, alpha=ALPHA)
    calibrated_d = perf_calib.data

    m0_expected =  _expected_m0(np.mean(calib_img.data), 1.0, 50, 0.82, alpha=ALPHA)
    np.testing.assert_allclose(calibrated_d, perf_img.data / m0_expected)



"""
ASL data preprocessing command line tool
"""
from __future__ import print_function

import sys

import scipy

import fsl.wrappers as fsl

from oxasl import Workspace, image
from oxasl.options import AslOptionParser, OptionCategory, IgnorableOptionGroup, GenericOptions

def preprocess(wsp):
    """
    Basic preprocessing of ASL data

    :param wsp: Workspace object

    Required workspace attributes
    -----------------------------

     - ``asldata`` : AslImage containing data to preprocess


    Optional workspace attributes
    -----------------------------

     - ``diff`` : If True, perform tag/control subtraction
     - ``reorder`` : If specified, reorder the data using this ordering string
     - ``mc`` : If True, perform motion correction using mcflirt
     - ``smooth`` : If True, perform smoothing
     - ``fwhm`` : If smooth=True, the full-width-half-maximum for the smoothing kernel
     - ``ref`` : If mc=True, a reference image for motion correction (default is to use the middle volume)

    Workspace attributes updated
    -----------------------------

     - ``asldata_preproc`` : preprocessed AslImage instance
    """
    wsp.log.write("ASL preprocessing...\n")

    wsp.asldata_preproc = wsp.asldata

    if wsp.diff:
        wsp.log.write("  - Tag-control subtraction\n")
        wsp.asldata_preproc = wsp.asldata_preproc.diff()

    if wsp.reorder:
        wsp.log.write("  - Re-ordering to %s\n" % wsp.reorder)
        if "l" in wsp.reorder.lower() and wsp.diff:
            wsp.reorder = wsp.reorder.replace("l", "")
        wsp.asldata_preproc = wsp.asldata_preproc.reorder(wsp.reorder)

    if wsp.mc:
        wsp.log.write("  - Motion correction\n")
        mcimg = fsl.mcflirt(wsp.asldata_preproc, cost="mutualinfo", out=fsl.LOAD, reffile=wsp.ref)["out"]
        wsp.asldata_preproc = wsp.asldata_preproc.derived(mcimg.data, suffix="_mc")

    if wsp.smooth:
        wsp.sigma = round(wsp.fwhm / 2.355, 2)
        wsp.log.write("  - Spatial smoothing with FWHM: %f (sigma=%f)\n" % (wsp.fwhm, wsp.sigma))
        smoothed = scipy.ndimage.gaussian_filter(wsp.asldata_preproc.data, sigma=wsp.sigma)
        wsp.asldata_preproc = wsp.asldata_preproc.derived(smoothed, suffix="_smooth")

    wsp.log.write("DONE\n\n")

class AslPreprocOptions(OptionCategory):
    """
    OptionCategory which contains options for preprocessing ASL data
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "preproc", **kwargs)

    def groups(self, parser):
        group = IgnorableOptionGroup(parser, "Preprocessing", ignore=self.ignore)
        group.add_option("--diff", help="Perform tag-control subtraction", action="store_true", default=False)
        group.add_option("--smooth", help="Spatially smooth data", action="store_true", default=False)
        group.add_option("--fwhm", help="FWHM for spatial filter kernel", type="float", default=6)
        group.add_option("--mc", help="Motion correct data", action="store_true", default=False)
        group.add_option("--ref", help="Optional reference image for motion correction", type="image", default=None)
        group.add_option("--reorder", help="Re-order data in specified order")
        return [group, ]

def main():
    """
    Entry point for command line tool
    """
    try:
        parser = AslOptionParser(usage="asl_preproc -i <filename> [options]")
        parser.add_category(GenericOptions(output_type="file", ignore=["mask"]))
        parser.add_category(image.AslImageOptions())
        parser.add_category(AslPreprocOptions())

        options, _ = parser.parse_args()
        if not options.output:
            options.output = "oxasl_preproc_out"
        wsp = Workspace(auto_asldata=True, **vars(options))

        print("")
        wsp.asldata.summary()
        preprocess(wsp)
        wsp.asldata_preproc.save(options.output)

    except ValueError as exc:
        sys.stderr.write("ERROR: " + str(exc) + "\n")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/bin/env python
"""
ASL_CALIB: Calibration for ASL data

Michael Chappell & Brad MacIntosh, FMRIB Image Analysis & Physics Groups

Copyright (c) 2008-2013 Univerisity of Oxford
"""

import sys
import os
import math
import traceback

import numpy as np
import scipy.ndimage

from fsl.data.image import Image
from fsl.data.atlases import AtlasRegistry

from oxasl import Workspace, struc, reg
from oxasl.image import summary
from oxasl.options import AslOptionParser, OptionCategory, IgnorableOptionGroup, GenericOptions
from oxasl.reporting import LightboxImage

def init(wsp):
    """ Initialize calibration sub-workspace """
    if wsp.calibration is None:
        wsp.sub("calibration")

def calculate_m0(wsp):
    """
    Calculate M0 value for use in calibration of perfusion images

    :param wsp: Workspace object

    Required workspace attributes
    -----------------------------

     - ``calib`` : Image containing voxelwise M0 map
     - ``calib_method`` : ``voxelwise`` or ``refregion``. Voxelwise calibration calibrates each voxel separately using
                          the corresponding voxel in the M0 map. Reference region calibration determines a single M0
                          value for the entire image by averaging over a region of the calibration image corresponding
                          to a known tissue type (e.g. CSF).

    Optional workspace attributes
    -----------------------------

     - ``calib_gain`` : Calibration gain (default 1.0)

    Additional optional and mandatory attributes may be required for different methods - see :ref get_m0_voxelwise:
    and :ref get_m0_refregion: functions for details.

    Workspace attributes updated
    -----------------------------

     - ``calibration.m0`` : M0 value, either a single value or an voxelwise Image
    """
    init(wsp)

    if wsp.calibration.m0 is None:
        wsp.log.write("\nCalibration - calculating M0\n")
        if wsp.calib_method in ("voxel", "voxelwise"):
            wsp.calibration.m0 = get_m0_voxelwise(wsp)
        elif wsp.calib_method in ("refregion", "single"):
            wsp.calibration.m0 = get_m0_refregion(wsp)
        elif wsp.calib_method == "wholebrain":
            wsp.calibration.m0 = get_m0_wholebrain(wsp)
        else:
            raise ValueError("Unknown calibration method: %s" % wsp.calib_method)

def calibrate(wsp, perf_img, multiplier=1.0, alpha=1.0, var=False):
    """
    Do calibration of a perfusion image from a calibration (M0) image

    :param wsp: Workspace object
    :param perf_img: Image containing perfusion data to calibrate
    :param multiplier: Scalar multiple to convert output to physical units
    :param alpha: Inversion efficiency
    :param var: If True, assume data represents variance rather than value

    :return: Image containing calibrated data

    Required workspace attributes
    -----------------------------

     - ``calibration.m0`` : M0 single value or voxelwise Image
    """
    if not perf_img:
        raise ValueError("Perfusion data cannot be None")
    if not wsp.calib:
        raise ValueError("No calibration data supplied")

    init(wsp)
    calculate_m0(wsp)
    wsp.log.write("\nCalibrating perfusion data: %s\n" % perf_img.name)
    m0 = wsp.calibration.m0
    if isinstance(m0, Image):
        m0 = m0.data

    if var:
        wsp.log.write(" - Treating data as variance - squaring M0 correction, multiplier and inversion efficiency\n")
        m0 = m0**2
        multiplier = multiplier**2
        alpha = alpha**2

    if isinstance(m0, np.ndarray):
        # If M0 is zero, make calibrated data zero
        calibrated = np.zeros(perf_img.shape)
        calibrated[m0 > 0] = perf_img.data[m0 > 0] / m0[m0 > 0]
    else:
        calibrated = perf_img.data / m0

    if alpha != 1.0:
        wsp.log.write(" - Using inversion efficiency correction: %f\n" % alpha)
        calibrated /= alpha

    if multiplier != 1.0:
        wsp.log.write(" - Using multiplier for physical units: %f\n" % multiplier)
        calibrated *= multiplier

    perf_calib = Image(calibrated, name=perf_img.name + "_calib", header=perf_img.header)
    return perf_calib

def get_m0_voxelwise(wsp):
    """
    Calculate M0 value using voxelwise calibration

    :param wsp: Workspace object
    :return: Image containing voxelwise M0 map

    Required workspace attributes
    -----------------------------

     - ``calib`` : Image containing calibration data

    Optional workspace attributes
    -----------------------------

     - ``calib_gain`` : Calibration gain (default 1.0)
     - ``tr`` : Sequence TR (s) (optional, if present used for short TR correction when tr<5)
     - ``t1t`` : Tissue T1 (optional, if present used for short TR correction)
     - ``pct`` : Partition coefficient used to convert M0 tissue into M0 arterial (default 0.9)
     - ``mask`` : Brain mask in calibration image space
     - ``calib_edgecorr`` : If True, and mask provided, apply edge correction
    """
    wsp.log.write(" - Doing voxelwise calibration\n")
    gain, pct = wsp.ifnone("calib_gain", 1), wsp.ifnone("pct", 0.9)
    wsp.log.write(" - Calibration gain: %f\n" % gain)

    # Calculate M0 value
    m0 = np.copy(wsp.calib.data).astype(np.float) * gain

    shorttr = 1
    if wsp.tr is not None and wsp.tr < 5:
        if wsp.t1t is not None:
            wsp.log.write(" - Correcting the calibration (M0) image for short TR (TR=%f, using T1 of tissue %f)\n" % (wsp.tr, wsp.t1t))
            shorttr = 1 / (1 - math.exp(-wsp.tr / wsp.t1t))
            m0 *= shorttr
        else:
            wsp.log.write("WARNING: tr < 5 (%f) but tissue T1 not provided so cannot apply short TR correction\n" % wsp.tr)

	# Include partiition co-effcient in M0 image to convert from M0 tissue to M0 arterial
    wsp.log.write(" - Using partition coefficient: %f\n" % pct)
    m0 /= pct

    if wsp.rois is not None and wsp.rois.mask is not None:
        if wsp.ifnone("calib_edgecorr", True):
            wsp.log.write(" - Doing edge correction\n")
            m0 = _edge_correct(m0, wsp.rois.mask)
        wsp.log.write(" - Masking M0 image")
        m0[wsp.rois.mask.data == 0] = 0

    m0img = Image(m0, header=wsp.calib.header)
    wsp.log.write(" - Mean M0: %f\n" % np.mean(m0))

    # Reporting
    page = wsp.report.page("m0")
    page.heading("Voxelwise M0 calculation")
    page.text("Voxelwise calibration calculates an M0 value for each voxel from the calibration image")
    page.heading("Correction factors", level=1)
    table = []
    table.append(["Calibration gain", "%.3g" % gain])
    table.append(["Blood/tissue partition coefficient", "%.3g" % pct])
    if wsp.tr is not None:
        table.append(["Sequence TR", "%.3g" % wsp.tr])
    if shorttr != 1:
        table.append(["Tissue T1", "%.3g" % wsp.t1t])
        table.append(["T1 Correction factor (for short TR)", "%.3g" % shorttr])
    table.append(["Overall correction factor", "%.3g" % (gain*shorttr/pct)])

    if wsp.rois is not None and wsp.rois.mask is not None:
        table.append(["Mean M0 (within mask)", "%.3g" % np.mean(m0[wsp.rois.mask.data > 0])])
        table.append(["Edge correction", "Enabled" if wsp.edgecorr else "Disabled"])
    else:
        table.append(["Mean M0", "%.3g" % np.mean(m0)])
    page.table(table)

    page.heading("M0 map", level=1)
    page.image("m0img", LightboxImage(m0img))

    return m0img

def _edge_correct(m0, brain_mask):
    """
    Correct for (partial volume) edge effects
    """
    brain_mask = brain_mask.data

    # Median smoothing
    m0 = scipy.ndimage.median_filter(m0, size=3)

    # Erode mask using 3x3x3 structuring element
    mask_ero = scipy.ndimage.morphology.binary_erosion(brain_mask, structure=np.ones([3, 3, 3]), border_value=1)
    m0[mask_ero == 0] = 0

    # Extrapolate remaining data to fit original mask
    # ASL_FILE works slicewise using a mean 5x5 filter on nonzero values, so we will do the same
    for z in range(m0.shape[2]):
        zslice = m0[..., z]
        zslice_extrap = scipy.ndimage.filters.generic_filter(zslice, _masked_mean, footprint=np.ones([5, 5]))
        m0[..., z] = zslice_extrap
    m0[brain_mask == 0] = 0

    return m0

def _masked_mean(vals):
    """
    Called by scipy.ndimage.filters.generic_filter

    :param vals: Values in the kernel (a 5x5 square patch centered on the voxel in this case)

    For nonzero voxels, returns the voxel value (i.e. the middle element of vals).
    For zero voxels, returns the mean of non zero voxels in the kernel
    """
    voxel_val = vals[int((len(vals)-1) / 2)]
    if voxel_val == 0:
        nonzero = vals[vals != 0]
        if np.any(nonzero):
            return np.mean(nonzero)
        else:
            return 0
    else:
        return voxel_val

def get_m0_wholebrain(wsp):
    """
    Get a whole-brain M0 value

    This calculates an M0 map for the whole brain using the T1, T2 and PC
    values for each tissue type with the contribution at each voxel weighted
    by the tissue partial volume

    :param wsp: Workspace object
    :return: Image containing voxelwise M0 map

    Required Workspace attributes
    -----------------------------

      - ``calib``     - Calibration Image in ASL native space
      - ``rois.mask`` - Brain mask Image in ASL native space
      - ``struc``     - Structural image
    """
    struc.segment(wsp)
    wsp.log.write("\n - Doing wholebrain region calibration\n")

    tr = wsp.ifnone("tr", 3.2)
    te = wsp.ifnone("te", 0)
    taq = wsp.ifnone("taq", 0)
    wsp.log.write(" - Using TE=%f, TR=%f, Readout time (TAQ)=%f\n" % (te, tr, taq))

    t2star = wsp.ifnone("t2star", False)
    if t2star:
        # From Petersen 2006 MRM 55(2):219-232 see discussion
        t2b = 50
    else:
        # From Lu et al. 2012 MRM 67:42-49 have 154ms at 3T during normoxia
        t2b = 150

    # Check the data and masks
    calib_data = np.copy(wsp.calib.data)
    if wsp.rois is not None and wsp.rois.mask is not None:
        brain_mask = wsp.rois.mask.data
    else:
        brain_mask = np.ones(wsp.calib.shape[:3])

    ### Sensitivity image calculation (if we have a sensitivity image)
    if wsp.sens:
        wsp.log.write(" - Using sensitivity image: %s\n" % wsp.sens.name)
        calib_data /= wsp.sens.data

    m0 = np.zeros(calib_data.shape, dtype=np.float)
    for tiss_type in ("wm", "gm", "csf"):
        pve_struc = getattr(wsp.structural, "%s_pv" % tiss_type)
        wsp.log.write(" - Transforming %s tissue PVE into ASL space\n" % tiss_type)
        pve = reg.struc2asl(wsp, pve_struc)
        t1r, t2r, t2rstar, pcr = tissue_defaults(tiss_type)
        if t2star:
            t2r = t2rstar
        t1_corr = 1 / (1 - math.exp(- (tr - taq) / t1r))
        t2_corr = 1 / math.exp(- te / t2r)
        wsp.log.write("Correction factors: T1: %f, T2 %f, PC: %f" % (t1_corr, t2_corr, pcr))
        tiss_m0 = calib_data * t1_corr * t2_corr / pcr
        wsp.log.write(" - Mean %s M0: %f (weighted by PV)\n" % (tiss_type, np.average(tiss_m0, weights=pve.data)))
        tiss_m0 *= pve.data
        setattr(wsp.calibration, "m0_img_%s" % tiss_type, Image(tiss_m0, header=wsp.calib.header))
        m0 += tiss_m0

    m0 = m0 * math.exp(- te / t2b)
    gain = wsp.ifnone("calib_gain", 1)
    wsp.log.write(" - Calibration gain: %f\n" % gain)
    if gain != 1:
        m0 = m0 * gain

    wsp.calibration.m0_img = Image(m0, header=wsp.calib.header)
    m0 = np.mean(m0[brain_mask != 0])
    wsp.log.write(" - M0 of brain: %f\n" % m0)

    # Reporting
    page = wsp.report.page("m0")
    page.heading("Whole-brain M0 calculation")
    page.text("Whole-brain calibration calculates an M0 value for each voxel from the calibration image based on partial volume estimates")
    page.text("- Calibration gain: %f" % gain)
    page.text(" - TR: %f" % tr)
    page.text(" - TE: %f" % te)

    page.heading("M0", level=1)
    page.text("Mean M0 value (within mask): %f" % m0)
    page.image("m0img", LightboxImage(wsp.calibration.m0_img))

    return float(m0)

def get_m0_refregion(wsp, mode="longtr"):
    """
    Do reference region calibration

    FIXME saturation recovery mode is not functional

    Required workspace attributes
    -----------------------------

     - ``calib`` : Image containing voxelwise M0 map
     - ``refmask`` : Reference region mask in calibration image space. Technically this is not required
                      as the brain mask will be used if not specified, however it is strongly recommended.

    Optional workspace attributes
    -----------------------------

     - ``calib_gain`` : Calibration gain (default 1.0)
     - ``tr`` : Sequence TR (s) (default 3.2)
     - ``te`` : Sequence TE (s) (default 0)
     - ``taq`` : Sequence TAQ (s) (default 0)
     - ``t2star`` : If True, correct for T2* rather than T2
     - ``t1r`` : Reference tissue T1 (default: see ``TISSUE_DEFAULTS``)
     - ``t2r`` : Reference tissue T2 (default: see ``TISSUE_DEFAULTS``)
     - ``t2rstar`` : Reference tissue T2* (default: see ``TISSUE_DEFAULTS``)
     - ``pcr`` : Reference tissue partition coefficient (default: see ``TISSUE_DEFAULTS``)
     - ``mask`` : Brain mask in calibration image space
    """
    wsp.log.write(" - Doing reference region calibration\n")
    gain = wsp.ifnone("calib_gain", 1)

    tr = wsp.ifnone("tr", 3.2)
    te = wsp.ifnone("te", 0)
    taq = wsp.ifnone("taq", 0)
    wsp.log.write(" - Acquisition: TE=%f, TR=%f, Readout time (TAQ)=%f\n" % (te, tr, taq))

    t2star = wsp.ifnone("t2star", False)
    if t2star:
        # From Petersen 2006 MRM 55(2):219-232 see discussion
        t2b = 50
    else:
        # From Lu et al. 2012 MRM 67:42-49 have 154ms at 3T during normoxia
        t2b = 150

    # Parameters for reference tissue type: T1, T2, T2*, partition coeffs, FAST seg ID
    # Partition coeffs based on Herscovitch and Raichle 1985 with a blood water density of 0.87
    # GM/WM T2* from Foucher 2011 JMRI 34:785-790
    wsp.tissref = wsp.ifnone("tissref", "csf")
    wsp.log.write(" - Using tissue reference type: %s\n" % wsp.tissref)

    t1r, t2r, t2rstar, pcr = tissue_defaults(wsp.tissref)
    if t2star:
        t2r = t2rstar

    # Command line override of default T1, T2, PC
    t1r_img, t2r_img = False, False

    if wsp.t1r is not None:
        t1r = wsp.t1r
        if isinstance(t1r, Image):
            wsp.log.write(" - Using T1 image for reference region: %s\n" % t1r.name)
            t1r_img = True
        else:
            wsp.log.write(" - Using user-specified T1r value: %f\n" % t1r)

    if wsp.t2r is not None:
        t2r = wsp.t2r
        if isinstance(t2r, Image):
            wsp.log.write(" - Using T2 image for reference region: %s\n" % t2r.name)
            t2r_img = True
        else:
            wsp.log.write(" - Using user-specified T2r value: %f\n" % t2r)

    if wsp.t2b is not None:
        t2b = wsp.t2b
        wsp.log.write(" - Using user-specified T2b value: %f\n" % t2b)

    if wsp.pcr is not None:
        pcr = wsp.pcr
        wsp.log.write(" - Using user-specified partition coefficient: %f\n" % pcr)

    if t1r is None:
        raise ValueError("T1 for reference tissue has not been set")
    if t2r is None:
        if te != 0:
            raise ValueError("T2 for reference tissue has not been set")
        else:
            t2r = 1.0
    if pcr is None:
        raise ValueError("Partition coefficient for reference tissue has not been set")

    wsp.log.write(" - T1r: %f; T2r: %f; T2b: %f; Part co-eff: %f\n" % (t1r, t2r, t2b, pcr))

    # Check the data and masks
    calib_data = np.copy(wsp.calib.data).astype(np.float)
    wsp.calibration.calib_img = wsp.calib
    if calib_data.ndim == 4:
        wsp.log.write(" - Taking mean across calibration images\n")
        calib_data = np.mean(calib_data, -1)

    if wsp.rois is not None and wsp.rois.mask is not None:
        brain_mask = wsp.rois.mask.data
    else:
        brain_mask = np.ones(wsp.calib.shape[:3])

    if wsp.refmask is not None:
        wsp.log.write(" - Using supplied reference tissue mask: %s\n" % wsp.refmask.name)
        wsp.calibration.refmask = Image(wsp.refmask.data.astype(np.int), header=wsp.refmask.header)
        wsp.calibration.refmask_trans = reg.calib2asl(wsp, wsp.calibration.refmask, mask=True)
        refmask = wsp.calibration.refmask_trans.data
    elif wsp.tissref.lower() in ("csf", "wm", "gm"):
        get_tissrefmask(wsp)
        refmask = wsp.calibration.refmask.data

    nonzero = np.count_nonzero(refmask)
    if nonzero < 1:
        raise ValueError("Reference mask does not contain any unmasked voxels")
    else:
        wsp.log.write(" - Number of voxels in tissue reference mask: %i\n" % nonzero)

    ### Sensitivity image calculation (if we have a sensitivity image)
    sens_corr = False
    if wsp.sens:
        wsp.log.write(" - Using sensitivity image: %s\n" % wsp.sens.name)
        sens_corr = True
        sens_data = wsp.sens.data

    wsp.log.write(" - MODE: %s\n" % mode)
    wsp.log.write(" - Calibration gain: %f\n" % gain)

    if mode == "longtr":
        if sens_corr:
            wsp.log.write(" - Applying sensitivity image\n")
            calib_data /= sens_data

        # Mask M0 map with tissue reference
        calib_data[refmask == 0] = 0

        # calcualte T1 of reference region (if a T1 image has been supplied)
        if t1r_img:
            t1r = np.mean(t1r.data[refmask != 0])
            wsp.log.write(" - Calculated T1 of reference tissue: %f\n" % t1r)

        # calcualte T2 of reference region (if a T2 image has been supplied)
        if t2r_img:
            t2r = np.mean(t2r.data[refmask != 0])
            wsp.log.write(" - Calculated T2 of reference tissue: %f\n" % t2r)

        # calculate M0_ref value
        mean_sig = np.mean(calib_data[refmask != 0])
        wsp.log.write(" - mean signal in reference tissue: %f\n" % mean_sig)
        t1_corr = 1 / (1 - math.exp(- (tr - taq) / t1r))
        wsp.log.write(" - T1 correction factor: %f\n" % t1_corr)

    elif mode == "satrecov":
        # Calibration image is control images and we want to do a saturation recovery fit
        # NB only do the fit in the CSF mask
        # FIXME this is not functional at the moment
        options = {
            "data" : wsp.calib,
            "mask" : refmask,
            "method" : "vb",
            "noise" : "white",
            "model" : "satrecov",
            "t1" : t1r,
        }

        #deal with TIs
        tis = wsp.ifnone("satrecov_tis", [])
        wsp.log.write(" - TIs: %s\n" % str(tis))
        for idx, ti in enumerate(tis):
            options["ti%i" % (idx+1)] = ti

        # Extra options for Look Locker
        if wsp.FA:
            options["FA"] = wsp.FA
        if wsp.calib_nphases:
            options["phases"] = wsp.calib_nphases
        if wsp.lfa:
            options["LFA"] = wsp.lfa

        # Extra sat recovery options
        if wsp.fixa:
            options["fixa"] = True

        # Do fabber within the tissue reference mask with a sensible T1 prior mean
        wsp.log.write(" - Running FABBER within reference tissue mask\n")
        from .wrappers import fabber
        fabber_result = fabber(options)
        mean_m0 = fabber_result["mean_M0t"]

        # Calculate M0 value - this is mean M0 of CSF at the TE of the sequence
        m0_value = np.mean(mean_m0.data[refmask != 0])

        wsp.log.write(" - M0 of reference tissue: %f\n" % m0_value)

        # Save useful results
        wsp.calibration.t1_ref = Image("%s/mean_T1t" % wsp.workdir)
        wsp.calibration.m0_ref = Image("%s/mean_M0t" % wsp.workdir)

        # Do fabber again within whole brain to get estimated T1 of tissue and FA correction (if LL)
        # (note that we do not apply sensitivity correction to the data here - thius is 'built-into' the M0t map)
        wsp.log.write(" - FABBER (again) within whole brain mask\n")
        options["mask"] = brain_mask
        fabber(options)
        # $fabber --data=$calib --mask=$bmask --output=$temp_calib/satrecovT --data-order=singlefile --model=satrecov --noise=white --method=vb $tislist $llopts $sropts

        # save useful results to specified output directory
        #imcp $temp_calib/satrecovT/mean_T1t $outdir/T1t
        #imcp $temp_calib/satrecovT/mean_M0t $outdir/M0t
        #if [ ! -z $lfa ]; then
        #imcp $temp_calib/satrecovT/mean_g $outdir/facorr
    else:
        raise ValueError("Unknown reference region mode: %s (Should be satrecov or longtr)" % mode)

    # Use equation to get the M0 value that is needed

    #  T2 correction
    t2_corr = math.exp(- te / t2b) / math.exp(- te / t2r)
    wsp.log.write(" - T2 correction factor: %f\n" % t2_corr)
    m0 = mean_sig * gain * t1_corr * t2_corr / pcr

    # Apply calibration to input image
    if sens_corr:
        # Apply sensitivity image
        # fslmaths $infile -div $temp_calib/sens $temp_calib/infile
        pass
    else:
        # imcp $infile $temp_calib/infile
        pass

    wsp.log.write(" - M0: %f\n" % m0)

    # Reporting
    page = wsp.report.page("m0")
    page.heading("Reference region M0 calculation")
    page.text("Reference region calibration calculates a single M0 value for a region of known tissue type")
    page.heading("Calculation details", level=1)
    page.table([
        ["Sequence TR (s)", tr],
        ["Sequence TE (ms)", te],
        ["T1 reference tissue (s)", t1r],
        ["T2 reference tissue (ms)", t2r],
        ["Blood T2 (ms)", t2b],
        ["Number of voxels in reference region", np.count_nonzero(refmask)],
        ["Mean signal in reference region", mean_sig],
        ["T1 correction factor", t1_corr],
        ["T2 correction factor", t2_corr],
        ["Partition coefficient", pcr],
        ["Calibration gain", gain],
        ["M0", m0],
    ])
    page.heading("Reference tissue mask", level=1)
    page.image("refmask", LightboxImage(wsp.calibration.refmask, bgimage=wsp.calib))

    return float(m0)

def get_tissrefmask(wsp):
    """
    Calculate a calibration reference mask for a particular known tissue type
    """
    struc.segment(wsp)

    page = wsp.report.page("auto_calib_mask")
    page.heading("Calibration reference region")
    page.text("Reference region was automatically generated for tissue type: %s" % wsp.tissref.upper())
    page.heading("Partial volume map for %s tissue (from structural segmentation)" % wsp.tissref.upper(), level=1)
    wsp.calibration.refpve = getattr(wsp.structural, "%s_pv" % wsp.tissref.lower())
    page.image("refpve", LightboxImage(wsp.calibration.refpve, bgimage=wsp.structural.brain))

    if wsp.tissref == "csf" and not wsp.csfmaskingoff:
        wsp.log.write(" - Doing automatic ventricle selection using standard atlas\n")
        # By deafult now we do FNRIT transformation of ventricle mask
        # FIXME disabled as not being used in ASL_CALIB at present
        #wsp.calibration.struc2stdfnirt = wsp.ifnone("stdmaskfnirt", True)

        # Select ventricles based on standard space atlas
        page.heading("Automatic ventricle selection", level=1)
        page.text("Standard space ventricles mask (from Harvard-Oxford atlas) eroded by 1 pixel")
        atlases = AtlasRegistry()
        atlases.rescanAtlases()
        atlas = atlases.loadAtlas("harvardoxford-subcortical", loadSummary=False, resolution=2)
        ventricles = ((atlas.data[..., 2] + atlas.data[..., 13]) > 0.1).astype(np.int)
        wsp.calibration.ventricles = Image(scipy.ndimage.binary_erosion(ventricles, structure=np.ones([3, 3, 3]), border_value=1).astype(np.int), header=atlas.header)
        std_img = Image(os.path.join(os.environ["FSLDIR"], "data", "standard", 'MNI152_T1_2mm_brain'))
        page.image("ventricles_std", LightboxImage(wsp.calibration.ventricles, bgimage=std_img))

        page.heading("Structural space ventricles mask", level=1)
        reg.reg_struc2std(wsp)
        # FIXME nearest neighbour interpolation?
        wsp.calibration.ventricles_struc = reg.std2struc(wsp, wsp.calibration.ventricles)
        page.text("This is the above image transformed into structural space. The transformation was obtained by registering the structural image to the standard brain image")
        page.image("ventricles_struc", LightboxImage(wsp.calibration.ventricles_struc, bgimage=wsp.structural.brain))

        wsp.log.write(" - Masking FAST output with standard space derived ventricle mask\n")
        wsp.calibration.refpve_pre_mask = wsp.calibration.refpve
        refpve_data = np.copy(wsp.calibration.refpve.data)
        refpve_data[wsp.calibration.ventricles_struc.data == 0] = 0
        wsp.calibration.refpve = Image(refpve_data, header=wsp.calibration.refpve.header)
        wsp.calibration.refpve_post = wsp.calibration.refpve

        page.heading("Structural space ventricles PVE", level=1)
        page.text("This is the CSF partial volume masked by the ventricles mask. It should select only the ventricles from the original partial volume image.")
        page.image("refpve_post", LightboxImage(wsp.calibration.refpve, bgimage=wsp.structural.brain))

    wsp.log.write(" - Transforming tissue reference mask into ASL space\n")
    # FIXME calibration image may not be in ASL space! Oxford_asl does not handle this currently
    wsp.calibration.refpve_calib = reg.struc2asl(wsp, wsp.calibration.refpve)
    #wsp.calibration.refpve_calib.data[wsp.calibration.refpve_calib.data < 0.001] = 0 # Better for display
    page.heading("Reference region in ASL space", level=1)
    page.text("Partial volume map")
    page.image("refpve_calib", LightboxImage(wsp.calibration.refpve_calib, bgimage=wsp.calib))

    # Threshold reference mask conservatively to select only reference tissue
    wsp.log.write(" - Thresholding reference mask\n")
    wsp.calibration.refmask = Image((wsp.calibration.refpve_calib.data > 0.9).astype(np.int), header=wsp.calibration.refpve_calib.header)

    page.text("Reference Mask (thresholded at 0.9")
    page.image("refmask", LightboxImage(wsp.calibration.refmask, bgimage=wsp.calib))

TISS_DEFAULTS = {
    "csf" : [4.3, 750, 400, 1.15],
    "gm" : [1.3, 100, 50, 0.98],
    "wm" : [1.0, 50, 60, 0.82],
}

def tissue_defaults(tiss_type=None):
    """
    Get default T1, T2, T2* and PC for different tissue types

    Parameters for reference tissue type: T1, T2, T2*, partition coeffs, FAST seg ID
    Partition coeffs based on Herscovitch and Raichle 1985 with a blood water density of 0.87
    GM/WM T2* from Foucher 2011 JMRI 34:785-790

    :return: If tiss_type given, tuple of T1, T2, T2* and PC for tissue type.
             Otherwise return dictionary of tissue type name (csf, wm, gm) to tuple
             of T1, T2, T2* and PC for all known types
    """
    if tiss_type is None:
        return TISS_DEFAULTS
    elif tiss_type.lower() in TISS_DEFAULTS:
        return TISS_DEFAULTS[tiss_type.lower()]
    else:
        raise ValueError("Invalid tissue type: %s" % tiss_type)

class CalibOptions(OptionCategory):
    """
    OptionCategory which contains options for preprocessing ASL data
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "calib", **kwargs)

    def groups(self, parser):
        groups = []
        group = IgnorableOptionGroup(parser, "Calibration", ignore=self.ignore)
        group.add_option("--calib", "-c", help="Calibration image", type="image")
        group.add_option("--perf", "-i", help="Perfusion image for calibration, in same image space as calibration image", type="image")
        group.add_option("--calib-method", "--cmethod", help="Calibration method: voxelwise or refregion")
        group.add_option("--calib-alpha", "--alpha", help="Inversion efficiency", type=float, default=None)
        group.add_option("--calib-gain", "--cgain", help="Relative gain between calibration and ASL data", type=float, default=1.0)
        group.add_option("--tr", help="TR used in calibration sequence (s)", type=float, default=3.2)
        groups.append(group)

        group = IgnorableOptionGroup(parser, "Voxelwise calibration", ignore=self.ignore)
        group.add_option("--pct", help="Tissue/arterial partition coefficiant", type=float, default=0.9)
        group.add_option("--t1t", help="T1 of tissue (s)", type=float, default=1.3)
        groups.append(group)

        group = IgnorableOptionGroup(parser, "Reference region calibration", ignore=self.ignore)
        group.add_option("--mode", help="Calibration mode (longtr or satrevoc)", default="longtr")
        group.add_option("--tissref", help="Tissue reference type (csf, wm, gm or none)", default="csf")
        group.add_option("--te", help="Sequence TE (ms)", type=float, default=0.0)
        group.add_option("--t1r", help="T1 of reference tissue (s) - defaults: csf 4.3, gm 1.3, wm 1.0", type=float, default=None)
        group.add_option("--t2r", help="T2/T2* of reference tissue (ms) - defaults T2/T2*: csf 750/400, gm 100/60,  wm 50/50", type=float, default=None)
        group.add_option("--t2b", help="T2/T2* of blood (ms) - default T2/T2*: 150/50)", type=float, default=None)
        group.add_option("--refmask", "--csf", help="Reference tissue mask in calibration image space", type="image")
        group.add_option("--t2star", action="store_true", default=False, help="Correct with T2* rather than T2 (alters the default T2 values)")
        group.add_option("--pcr", help="Reference tissue partition coefficiant (defaults csf 1.15, gm 0.98,  wm 0.82)", type=float, default=None)
        groups.append(group)

        group = IgnorableOptionGroup(parser, "longtr mode (calibration image is a control image with a long TR)", ignore=self.ignore)
        groups.append(group)

        group = IgnorableOptionGroup(parser, "satrecov mode (calibration image is a sequnce of control images at various TIs)", ignore=self.ignore)
        group.add_option("--tis", help="Comma separated list of inversion times, e.group. --tis 0.2,0.4,0.6")
        group.add_option("--fa", help="Flip angle (in degrees) for Look-Locker readouts", type=float)
        group.add_option("--lfa", help="Lower flip angle (in degrees) for dual FA calibration", type=float)
        group.add_option("--calib-nphases", help="Number of phases (repetitions) of higher FA", type=int)
        group.add_option("--fixa", action="store_true", default=False, help="Fix the saturation efficiency to 100% (useful if you have a low number of samples)")
        groups.append(group)

        #group = IgnorableOptionGroup(parser, "Coil sensitivity correction, either using existing sensitivity image or reference images collected using same parameters", ignore=self.ignore)
        #groups.append(group)

        #group = IgnorableOptionGroup(parser, "CSF masking options (only for --tissref csf)", ignore=self.ignore)
        #group.add_option("--csfmaskingoff", action="store_true", default=False, help="Turn off the ventricle masking, reference is based on segmentation only.")
        #group.add_option("--str2std", help="Structural to MNI152 linear registration (.mat)")
        #group.add_option("--warp", help="Structural to MNI152 non-linear registration (warp)")

        return groups

def main():
    """
    Entry point for oxasl_calib command line program
    """

    debug = False
    try:
        parser = AslOptionParser(usage="oxasl_calib -i <perfusion image> -c <calibration image> --calib-method <voxelwise|refregion> -o <output filename> [options]")
        parser.add_category(CalibOptions())
        parser.add_category(GenericOptions(output_type="file"))
        options, _ = parser.parse_args(sys.argv)

        if not options.perf:
            sys.stderr.write("Perfusion input file not specified\n")
            parser.print_help()
            sys.exit(1)

        if not options.calib:
            sys.stderr.write("Calibration input file not specified\n")
            parser.print_help()
            sys.exit(1)

        wsp = Workspace(**vars(options))

        summary(wsp.perf)
        summary(wsp.calib)

        if wsp.output is None:
            wsp.output = "%s_calib" % wsp.perf.name

        calibrated_img = calibrate(wsp, wsp.perf)
        summary(calibrated_img)
        calibrated_img.save(wsp.output)

    except ValueError as exc:
        sys.stderr.write("ERROR: " + str(exc) + "\n")
        if debug:
            traceback.print_exc()
        sys.exit(1)

if __name__ == "__main__":
    main()



"""
Wrapper for avscale command
"""

import fsl.utils.assertions as asrt
from fsl.wrappers import wrapperutils  as wutils

def extract_avscale_output(fn):
    def _wrapper(*args, **kwargs):
        capture = six.StringIO()
        log = kwargs.pop("log", {})
        old_stdout = log.get("stdout", None)
        log["stdout"] = capture
        result = fn(*args, log=log, **kwargs)
        print(result)
        if old_stdout is not None:
            old_stdout.write(capture.getvalue())
        for line in capture.getvalue().splitlines():
            print(line)
        return capture.getvalue()
        
    return _wrapper

@extract_avscale_output
@wutils.fileOrArray('matrixfile')
@wutils.fslwrapper
def avscale(matrixfile):
    """Wrapper for the ``avscale`` command.
    
    Required options:
    
    :arg matrixfile:    FLIRT transformation matrix 
    """
    cmd = ['avscale', '--allparams', matrixfile]
    return cmd



"""
oxasl.gui.widgets.py

Useful wx widgets for building the OXASL GUI

Copyright (c) 2019 University of Oxford
"""
import os
import colorsys

import wx
import wx.grid

import matplotlib
matplotlib.use('WXAgg')
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.figure import Figure

from fsl.data.image import Image

class OptionError(RuntimeError):
    """
    Can be thrown by tab page if the current configuration
    does not allow the pipeline to be run. Should contain
    a user-friendly message explaining the problem
    (e.g. 'Output directory not specified'
    """

class TabPage(wx.Panel):
    """
    Shared methods used by the various tab pages in the GUI
    """
    def __init__(self, parent, title, idx, n, name=None):
        wx.Panel.__init__(self, parent=parent, id=wx.ID_ANY)
        self.notebook = parent
        self.idx = idx
        self.n = n
        self.sizer = wx.GridBagSizer(vgap=5, hgap=5)
        self.row = 0
        self.title = title
        if name is None:
            self.name = title.lower()
        else:
            self.name = name

    def options(self):
        """
        :return: Dictionary of OXASL options from this tab page
        """
        return {}

    def next_prev(self):
        """
        Add next/previous buttons
        """
        if self.idx < self.n-1:
            self.next_btn = wx.Button(self, label="Next", id=wx.ID_FORWARD)
            self.next_btn.Bind(wx.EVT_BUTTON, self._next)
        else:
            self.next_btn = wx.StaticText(self, label="")

        if self.idx > 0:
            self.prev_btn = wx.Button(self, label="Previous", id=wx.ID_BACKWARD)
            self.prev_btn.Bind(wx.EVT_BUTTON, self._prev)
        else:
            self.prev_btn = wx.StaticText(self, label="")

        self.pack(" ")
        self.sizer.AddGrowableRow(self.row-1, 1)
        self.sizer.Add(self.prev_btn, pos=(self.row, 0), border=10, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_LEFT)
        self.sizer.Add(wx.StaticText(self, label=""), pos=(self.row, 1), border=10, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_LEFT)
        self.sizer.Add(wx.StaticText(self, label=""), pos=(self.row, 2), border=10, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_LEFT)
        self.sizer.Add(self.next_btn, pos=(self.row, 3), border=10, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_RIGHT)

    def _next(self, _):
        """
        Move to the next tab in the notebook
        """
        self.notebook.SetSelection(self.idx+1)

    def _prev(self, _):
        """
        Move to the previous tab in the notebook
        """
        self.notebook.SetSelection(self.idx-1)

    def pack(self, label, *widgets, **kwargs):
        """
        Add a horizontal line to the tab with a label and series of widgets

        If label is empty, first widget is used instead (usually to provide a checkbox)
        """
        col = 0
        border = kwargs.get("border", 10)
        font = self.GetFont()
        if "size" in kwargs:
            font.SetPointSize(kwargs["size"])
        if kwargs.get("bold", False):
            font.SetWeight(wx.BOLD)

        if label != "":
            text = wx.StaticText(self, label=label)
            text.SetFont(font)
            self.sizer.Add(text, pos=(self.row, col), border=border, flag=wx.ALIGN_CENTRE_VERTICAL | wx.LEFT)
            col += 1
        else:
            text = None

        for w in widgets:
            span = (1, 1)
            w.label = text
            if hasattr(w, "span"):
                span = (1, w.span)
            w.SetFont(font)
            w.Enable(col == 0 or kwargs.get("enable", True))
            self.sizer.Add(w, pos=(self.row, col), border=border, flag=wx.ALIGN_CENTRE_VERTICAL | wx.EXPAND | wx.LEFT, span=span)
            col += span[1]
        self.row += 1

    def file_picker(self, label, pick_dir=False, handler=None, optional=False, initial_on=False, pack=True, **kwargs):
        """
        Add a file picker to the tab
        """
        if not handler:
            handler = self._changed
        if pick_dir:
            picker = wx.DirPickerCtrl(self, style=wx.DIRP_USE_TEXTCTRL)
            picker.Bind(wx.EVT_DIRPICKER_CHANGED, handler)
        else:
            picker = wx.FilePickerCtrl(self)
            picker.Bind(wx.EVT_FILEPICKER_CHANGED, handler)
        picker.span = 2
        if optional:
            cb = wx.CheckBox(self, label=label)
            cb.SetValue(initial_on)
            cb.Bind(wx.EVT_CHECKBOX, handler)
            picker.checkbox = cb
            if pack:
                self.pack("", cb, picker, enable=initial_on, **kwargs)
        elif pack:
            self.pack(label, picker, **kwargs)

        return picker

    def choice(self, label, choices, initial=0, optional=False, initial_on=False, handler=None, pack=True, **kwargs):
        """
        Add a widget to choose from a fixed set of options
        """
        if not handler:
            handler = self._changed
        ch = wx.Choice(self, choices=choices)
        ch.SetSelection(initial)
        ch.Bind(wx.EVT_CHOICE, handler)
        if optional:
            cb = wx.CheckBox(self, label=label)
            cb.SetValue(initial_on)
            cb.Bind(wx.EVT_CHECKBOX, self._changed)
            ch.checkbox = cb
            if pack:
                self.pack("", cb, ch, enable=initial_on, **kwargs)
        elif pack:
            self.pack(label, ch, **kwargs)
        return ch

    def number(self, label, handler=None, **kwargs):
        """
        Add a widget to choose a floating point number
        """
        if not handler:
            handler = self._changed
        num = NumberChooser(self, changed_handler=handler, **kwargs)
        num.span = 2
        self.pack(label, num, **kwargs)
        return num

    def integer(self, label, handler=None, pack=True, minval=1, maxval=100, **kwargs):
        """
        Add a widget to choose an integer
        """
        if not handler:
            handler = self._changed
        spin = wx.SpinCtrl(self, min=minval, max=maxval, **kwargs)
        spin.SetValue(kwargs.get("initial", 0))
        spin.Bind(wx.EVT_SPINCTRL, handler)
        if pack:
            self.pack(label, spin)
        return spin

    def checkbox(self, label, initial=False, handler=None, **kwargs):
        """
        Add a simple on/off option
        """
        cb = wx.CheckBox(self, label=label)
        cb.span = 2
        cb.SetValue(initial)
        if handler:
            cb.Bind(wx.EVT_CHECKBOX, handler)
        else: cb.Bind(wx.EVT_CHECKBOX, self._changed)
        self.pack("", cb, **kwargs)
        return cb

    def section(self, label):
        """
        Add a section heading
        """
        self.pack(label, bold=True)

    def _changed(self, _):
        self.update()

    def update(self):
        """
        Update the run module, i.e. when options have changed
        """
        if hasattr(self, "run"):
            self.run.update()
            if hasattr(self, "preview"):
                self.preview.run = self.run

    def image(self, label, fname):
        """
        Check if a specified filename is a valid Nifti image
        """
        if not os.path.exists(fname):
            raise OptionError("%s - no such file or directory" % label)
        try:
            return Image(fname)
        except:
            raise OptionError("%s - invalid image file" % label)

class NumberChooser(wx.Panel):
    """
    Widget for choosing a floating point number
    """

    def __init__(self, parent, label=None, minval=0, maxval=1, initial=0.5, step=0.1, digits=2, changed_handler=None):
        super(NumberChooser, self).__init__(parent)
        self.minval, self.orig_min, self.maxval, self.orig_max = minval, minval, maxval, maxval
        self.handler = changed_handler
        self.hbox = wx.BoxSizer(wx.HORIZONTAL)
        if label is not None:
            self.label = wx.StaticText(self, label=label)
            self.hbox.Add(self.label, proportion=0, flag=wx.ALIGN_CENTRE_VERTICAL)
        # Set a very large maximum as we want to let the user override the default range
        #self.spin = wx.SpinCtrl(self, min=0, max=100000, initial=initial)
        #self.spin.Bind(wx.EVT_SPINCTRL, self._spin_changed)
        self.spin = wx.SpinCtrlDouble(self, min=0, max=100000, inc=step, initial=initial)
        self.spin.SetDigits(digits)
        self.spin.Bind(wx.EVT_SPINCTRLDOUBLE, self._spin_changed)
        self.slider = wx.Slider(self, value=initial, minValue=0, maxValue=100)
        self.slider.SetValue(100*(initial-self.minval)/(self.maxval-self.minval))
        self.slider.Bind(wx.EVT_SLIDER, self._slider_changed)
        self.hbox.Add(self.slider, proportion=1, flag=wx.EXPAND | wx.ALIGN_CENTRE_VERTICAL)
        self.hbox.Add(self.spin, proportion=0, flag=wx.EXPAND | wx.ALIGN_CENTRE_VERTICAL)
        self.SetSizer(self.hbox)

    def GetValue(self):
        """
        :return: numeric value selected
        """
        return self.spin.GetValue()

    def SetValue(self, val):
        """
        Set the numeric value displayed
        """
        self.spin.SetValue(val)
        self.slider.SetValue(100*(val-self.minval)/(self.maxval-self.minval))

    def _slider_changed(self, event):
        v = event.GetInt()
        val = self.minval + (self.maxval-self.minval)*float(v)/100
        self.spin.SetValue(val)
        if self.handler:
            self.handler(event)
        event.Skip()

    def _spin_changed(self, event):
        """ If user sets the spin outside the current range, update the slider range
        to match. However if they go back inside the current range, revert to this for
        the slider"""
        val = event.GetValue()
        if val < self.minval:
            self.minval = val
        elif val > self.orig_min:
            self.minval = self.orig_min
        if val > self.maxval:
            self.maxval = val
        elif val < self.orig_max:
            self.maxval = self.orig_max
        self.slider.SetValue(100*(val-self.minval)/(self.maxval-self.minval))
        if self.handler:
            self.handler(event)
        event.Skip()

class NumberList(wx.grid.Grid):
    """
    Widget for specifying a list of numbers
    """

    def __init__(self, parent, n, default=1.8):
        super(NumberList, self).__init__(parent, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, 0)
        self.n = 0
        self.default = default
        self.CreateGrid(1, 0)
        self.SetRowLabelSize(0)
        self.SetColLabelSize(0)
        self.set_size(n)
        self.Bind(wx.EVT_SIZE, self._on_size)

    def GetValues(self):
        """
        :return: Sequence of values entered
        """
        try:
            return [float(self.GetCellValue(0, c)) for c in range(self.n)]
        except ValueError:
            raise RuntimeError("Non-numeric values in number list")

    def set_size(self, n):
        """
        Set the number of values in the list
        """
        if self.n == 0:
            default = self.default
        else: default = self.GetCellValue(0, self.n-1)
        if n > self.n:
            self.AppendCols(n - self.n)
            for c in range(self.n, n):
                self.SetCellValue(0, c, str(default))
        elif n < self.n:
            self.DeleteCols(n, self.n-n)
        self.n = n
        self._resize_cols()

    def _resize_cols(self):
        w, _ = self.GetClientSize()
        cw = w / self.n
        for i in range(self.n):
            self.SetColSize(i, cw)

    def _on_size(self, event):
        event.Skip()
        self._resize_cols()

class PreviewPanel(wx.Panel):
    """
    Panel providing a simple image preview for the output of ASL_FILE.

    Used so user can check their choice of data grouping/ordering looks right
    """
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, size=wx.Size(300, 600))
        self.data = None
        self.run = None
        self.slice = -1
        self.nslices = 1
        self.view = 0
        self.figure = Figure(figsize=(3.5, 3.5), dpi=100, facecolor='black')
        self.axes = self.figure.add_subplot(111, facecolor='black')
        self.axes.get_xaxis().set_ticklabels([])
        self.axes.get_yaxis().set_ticklabels([])
        self.canvas = FigureCanvas(self, -1, self.figure)
        self.canvas.mpl_connect('scroll_event', self.scroll)
        self.canvas.mpl_connect('button_press_event', self.view_change)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        font = self.GetFont()
        font.SetWeight(wx.BOLD)
        text = wx.StaticText(self, label="Data preview - perfusion weighted image")
        text.SetFont(font)
        self.sizer.AddSpacer(10)
        self.sizer.Add(text, 0)
        self.sizer.Add(self.canvas, 2, border=5, flag=wx.EXPAND | wx.ALL)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        hbox.Add(wx.StaticText(self, label="Use scroll wheel to change slice, double click to change view"), 0, flag=wx.ALIGN_CENTRE_VERTICAL)
        self.update_btn = wx.Button(self, label="Update")
        self.update_btn.Bind(wx.EVT_BUTTON, self._changed)
        hbox.Add(self.update_btn)
        self.sizer.Add(hbox)

        self.sizer.AddSpacer(10)
        text = wx.StaticText(self, label="Data order preview")
        text.SetFont(font)
        self.sizer.Add(text, 0)
        self.order_preview = AslDataPreview(self, 1, 1, True, "trp", True)
        self.sizer.Add(self.order_preview, 2, wx.EXPAND)
        self.SetSizer(self.sizer)
        self.Layout()

    def _changed(self, _):
        self.update()

    def update(self):
        """
        Update the preview. This is called explicitly when the user clicks the update
        button as it may be slow
        """
        self.data = None
        if self.run is not None:
            self.data = self.run.get_preview_data()

        if self.data is not None:
            self.view = 0
            self._init_view()
        self.redraw()

    def _init_view(self):
        self.nslices = self.data.shape[2-self.view]
        self.slice = int(self.nslices / 2)
        self.redraw()

    def redraw(self):
        """
        Redraw the preview image
        """
        self.axes.clear()
        if self.data is None:
            return
        if self.view == 0:
            sl = self.data[:, :, self.slice]
        elif self.view == 1:
            sl = self.data[:, self.slice, :]
        else:
            sl = self.data[self.slice, :, :]

        i = self.axes.imshow(sl.T, interpolation="nearest", vmin=sl.min(), vmax=sl.max())
        self.axes.set_ylim(self.axes.get_ylim()[::-1])
        i.set_cmap("gray")
        self.Layout()
        self.canvas.draw()
        self.canvas.Refresh()

    def view_change(self, event):
        """
        Called on mouse click event. Double click changes the view direction and redraws
        """
        if self.data is None:
            return
        if event.dblclick:
            self.view = (self.view + 1) % 3
            self._init_view()
            self.redraw()

    def scroll(self, event):
        """
        Called on mouse scroll wheel to move through the slices in the current view
        """
        if event.button == "up":
            if self.slice != self.nslices-1:
                self.slice += 1
        else:
            if self.slice != 0:
                self.slice -= 1
        self.redraw()

class AslDataPreview(wx.Panel):
    """
    Widget to display a preview of the data ordering selected (i.e. how the volumes in a 4D
    dataset map to TIs, repeats and tag/control pairs)
    """

    def __init__(self, parent, n_tis, n_repeats, tc_pairs, order, tagfirst):
        wx.Panel.__init__(self, parent, size=wx.Size(300, 300))
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.Bind(wx.EVT_SIZE, self._on_size)
        self.Bind(wx.EVT_PAINT, self._on_paint)
        self.n_tis = n_tis
        self.n_repeats = int(n_repeats)
        self.tc_pairs = tc_pairs
        self.tagfirst = tagfirst
        self.order = order
        self.tis_name = "PLDs"

    def _on_size(self, event):
        event.Skip()
        self.Refresh()

    def _get_col(self, pos, ti):
        if ti:
            h = 170.0/255
        else:
            h = 90.0/255
        s, v = 0.5, 0.95 - pos/2
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        return wx.Colour(int(r*255), int(g*255), int(b*255))

    def _on_paint(self, _):
        w, h = self.GetClientSize()
        N = self.n_tis * int(self.n_repeats)
        if self.tc_pairs:
            N *= 2
        dc = wx.AutoBufferedPaintDC(self)
        dc.Clear()

        leg_width = (w-100)/4
        leg_start = 50

        dc.SetBrush(wx.TRANSPARENT_BRUSH)
        rect = wx.Rect(leg_start, 20, leg_width/4, 20)
        dc.GradientFillLinear(rect, self._get_col(0, True), self._get_col(1.0, True), wx.EAST)
        dc.DrawRectangle(*rect.Get())
        dc.DrawText(self.tis_name, leg_start+leg_width/3, 20)

        rect = wx.Rect(leg_start+leg_width, 20, leg_width/4, 20)
        dc.GradientFillLinear(rect, self._get_col(0, False), self._get_col(1.0, False), wx.EAST)
        dc.DrawRectangle(*rect.Get())
        dc.DrawText("Repeats", leg_start+4*leg_width/3, 20)

        if self.tc_pairs:
            dc.SetBrush(wx.TRANSPARENT_BRUSH)
            dc.DrawRectangle(leg_start+leg_width*2, 20, leg_width/4, 20)
            dc.DrawText("Label", leg_start+7*leg_width/3, 20)

            b = wx.Brush('black', wx.BDIAGONAL_HATCH)
            dc.SetBrush(b)
            dc.DrawRectangle(leg_start+leg_width*3, 20, leg_width/4, 20)
            dc.DrawText("Control", leg_start+10*leg_width/3, 20)

        dc.DrawRectangle(50, 50, w-100, h-100)
        dc.DrawRectangle(50, 50, w-100, h-100)
        dc.DrawText("0", 50, h-50)
        dc.DrawText(str(N), w-50, h-50)

        seq = [1,]
        for t in self.order[::-1]:
            temp = seq
            seq = []
            for i in temp:
                if t == "t":
                    seq += [i,] * self.n_tis
                elif t == "l":
                    if not self.tc_pairs:
                        seq.append(i)
                    elif self.tagfirst:
                        seq.append(i)
                        seq.append(i+1)
                    else:
                        seq.append(i+1)
                        seq.append(i)
                elif t == "r":
                    seq.append(i)
                    seq += [i+2,] * (int(self.n_repeats) - 1)

        tistart = -1
        ti_sep = 1
        for idx, s in enumerate(seq):
            if s == 1 and tistart < 0:
                tistart = idx
            elif s == 1:
                ti_sep = idx - tistart
                break

        bwidth = float(w - 100) / N
        x = 50
        pos = 0.0
        ti = 0
        d = 1.0/self.n_tis
        for idx, s in enumerate(seq):
            dc.SetPen(wx.TRANSPARENT_PEN)
            b = wx.Brush(self._get_col(pos, s in (1, 2)), wx.SOLID)
            dc.SetBrush(b)
            dc.DrawRectangle(int(x), 50, int(bwidth+1), h-100)

            if s in (2, 4):
                b = wx.Brush('black', wx.BDIAGONAL_HATCH)
                dc.SetBrush(b)
                dc.DrawRectangle(int(x), 50, int(bwidth+1), h-100)

            dc.SetPen(wx.Pen('black'))
            dc.DrawLine(int(x), 50, int(x), h-50)
            x += bwidth
            if (idx+1) % ti_sep == 0:
                pos += d
                ti += 1
                if ti == self.n_tis:
                    pos = 0
                    ti = 0



"""
Structural data module for ASL

Copyright (c) 2008-2018 University of Oxford
"""
import os
import glob

import numpy as np

import fsl.wrappers as fsl
from fsl.data.image import Image
from fsl.utils.path import PathError

from oxasl.options import OptionCategory, IgnorableOptionGroup
from oxasl.reporting import LightboxImage

class StructuralImageOptions(OptionCategory):
    """
    OptionGroup which contains options for describing a structural image
    """

    def __init__(self, title="Structural image", **kwargs):
        OptionCategory.__init__(self, "struc", **kwargs)
        self.title = title

    def groups(self, parser):
        group = IgnorableOptionGroup(parser, self.title, ignore=self.ignore)
        group.add_option("--struc", "-s", help="Structural image", type="image", default=None)
        group.add_option("--struc-brain", "--sbet", "--struc-bet", type="image", help="Structural image (brain extracted)", default=None)
        group.add_option("--struc2asl", help="Structural->ASL transformation matrix", default=None)
        group.add_option("--asl2struc", help="ASL->Structural transformation matrix", default=None)
        group.add_option("--wm-seg", help="White matter segmentation of structural image", type="image", default=None)
        group.add_option("--gm-seg", help="Grey matter segmentation of structural image", type="image", default=None)
        group.add_option("--csf-seg", help="CSF segmentation of structural image", type="image", default=None)
        group.add_option("--fslanat", help="FSL_ANAT output directory for structural information", default=None)
        group.add_option("--fastsrc", help="Images from a FAST segmentation - if not set FAST will be run on structural image")
        group.add_option("--struc2std", help="Structural to MNI152 linear registration (.mat)")
        group.add_option("--struc2std-warp", help="Structural to MNI152 non-linear registration (warp)")

        return [group, ]

def init(wsp):
    """
    Do initialization on supplied structural data - copy relevant image and do brain extraction

    FIXME copy across all supplied structural data
    """
    if wsp.structural is not None:
        return

    wsp.log.write("\nInitialising structural data\n")
    wsp.sub("structural")

    if wsp.fslanat:
        wsp.log.write(" - Using FSL_ANAT output directory for structural data: %s\n" % wsp.fslanat)
        biascorr = os.path.join(wsp.fslanat, "T1_biascorr")
        biascorr_brain = os.path.join(wsp.fslanat, "T1_biascorr_brain")
        if glob.glob(biascorr + ".*") and glob.glob(biascorr_brain + ".*"):
            wsp.log.write(" - Using bias-corrected structural images\n")
            wsp.structural.struc = Image(biascorr)
            wsp.structural.brain = Image(biascorr_brain)
        else:
            wsp.log.write(" - Using non bias-corrected structural images\n")
            wsp.structural.struc = Image(os.path.join(wsp.fslanat, "T1"))
            wsp.structural.brain = Image(os.path.join(wsp.fslanat, "T1_brain"))
    elif wsp.struc:
        wsp.log.write(" - Using structural image provided by user: %s\n" % wsp.struc.name)
        wsp.structural.struc = wsp.struc
        if wsp.wm_seg is not None:
            wsp.structural.wm_seg = wsp.wm_seg # FIXME user override segmentation

    #elif wsp.structural.struc_lores
    #    wsp.log.write("Low-resolution tructural image: %s\n" % wsp.structural.struc_lores.name)
    else:
        wsp.log.write(" - No structural data supplied - output will be native space only\n")

    if wsp.structural.struc is not None and wsp.structural.brain is None:
        wsp.log.write(" - Brain-extracting structural image\n")
        bet_result = fsl.bet(wsp.structural.struc, output=fsl.LOAD, seg=True, mask=True, log=wsp.fsllog)
        wsp.structural.brain = bet_result["output"]
        #wsp.structural.brain_mask = bet_result["output_mask"]

    if wsp.structural.brain is not None and wsp.structural.brain_mask is None:
        # FIXME - for now get the mask by binarising the brain image but gives slightly
        # different results compared to using the mask returned by BET
        wsp.structural.brain_mask = Image((wsp.structural.brain.data != 0).astype(np.int), header=wsp.structural.struc.header)

    if wsp.structural.struc is not None:
        segment(wsp)

def segment(wsp):
    """
    Segment the structural image
    """
    init(wsp)
    if None in (wsp.structural.wm_seg, wsp.structural.gm_seg, wsp.structural.csf_seg):
        init(wsp)
        page = wsp.report.page("seg")
        page.heading("Segmentation of structural image")

        wsp.log.write("\nGetting structural segmentation\n")
        if wsp.fslanat:
            wsp.log.write(" - Using FSL_ANAT output\n")
            page.text("Segmentation taken from FSL_ANAT output at ``%s``" % wsp.fslanat)
            wsp.structural.csf_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_0"))
            wsp.structural.gm_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_1"))
            wsp.structural.wm_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_2"))

            try:
                wsp.structural.bias = Image(os.path.join(wsp.fslanat, "T1_fast_bias"))
                wsp.log.write(" - Bias field extracted sucessfully\n")
            except PathError:
                wsp.log.write(" - No bias field found")
        elif wsp.fastsrc:
            # FIXME should be possible to implement this
            raise NotImplementedError("Specifying FAST output directory")
            #img = os.path.split(wsp.fastsrc)[1]
            #wsp.structural.csf_pv = os.path.join(wsp.fastsrc, "%s_pve_0" % img)
            #wsp.structural.gm_pv = os.path.join(wsp.fastsrc, "%s_pve_1" % img)
            #wsp.structural.wm_pv = os.path.join(wsp.fastsrc, "%s_pve_2" % img)
        elif wsp.structural.struc:
            wsp.log.write(" - Running FAST\n")
            page.text("FAST run to segment structural image")
            fast_result = fsl.fast(wsp.structural.brain, out=fsl.LOAD, log=wsp.fsllog)
            wsp.structural.csf_pv = fast_result["out_pve_0"]
            wsp.structural.gm_pv = fast_result["out_pve_1"]
            wsp.structural.wm_pv = fast_result["out_pve_2"]
            #wsp.bias_struc = fast_result["fast_bias"]
        else:
            raise ValueError("No structural data provided - cannot segment")

        wsp.structural.csf_seg = Image((wsp.structural.csf_pv.data > 0.5).astype(np.int), header=wsp.structural.struc.header)
        wsp.structural.gm_seg = Image((wsp.structural.gm_pv.data > 0.5).astype(np.int), header=wsp.structural.struc.header)
        wsp.structural.wm_seg = Image((wsp.structural.wm_pv.data > 0.5).astype(np.int), header=wsp.structural.struc.header)

        page.heading("Segmentation image", level=1)
        page.text("CSF partial volume")
        page.image("csf_pv", LightboxImage(wsp.structural.csf_pv, bgimage=wsp.structural.brain))
        page.text("Grey matter partial volume")
        page.image("gm_pv", LightboxImage(wsp.structural.gm_pv, bgimage=wsp.structural.brain))
        page.text("White matter partial volume")
        page.image("wm_pv", LightboxImage(wsp.structural.wm_pv, bgimage=wsp.structural.brain))



#!/usr/bin/env python
"""
Simple wxpython based GUI front-end to OXASL pipeline

Currently this does not use the GUI system from the FSL python libraries.
Possible improvements would include:

    - Use props library to hold run options and use the signalling mechanisms to communicate
    values. The built-in widget builder does not seem to be flexible enough however.
    - Use fsleyes embedded widget as the preview for a nicer, more interactive data preview

Requirements (beyond OXASL requirements):
    - wxpython
    - matplotlib

Copyright (c) 2019 University of Oxford
"""

import os

import wx

from oxasl.gui.analysis_tab import AslAnalysis
from oxasl.gui.structure_tab import StructureTab
from oxasl.gui.calib_tab import AslCalibration
from oxasl.gui.input_tab import AslInputOptions
from oxasl.gui.dist_corr_tab import AslDistCorr
from oxasl.gui.widgets import PreviewPanel
from oxasl.gui.run_box import AslRun

class AslGui(wx.Frame):
    """
    Main GUI window
    """

    def __init__(self):
        wx.Frame.__init__(self, None, title="OXASL", size=(1200, 700), style=wx.DEFAULT_FRAME_STYLE)
        main_panel = wx.Panel(self)
        main_vsizer = wx.BoxSizer(wx.VERTICAL)

        banner = wx.Panel(main_panel, size=(-1, 80))
        banner.SetBackgroundColour((57, 71, 121))
        banner_fname = os.path.join(os.path.abspath(os.path.dirname(__file__)), "banner.png")
        wx.StaticBitmap(banner, -1, wx.Bitmap(banner_fname, wx.BITMAP_TYPE_ANY))
        main_vsizer.Add(banner, 0, wx.EXPAND)

        hpanel = wx.Panel(main_panel)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        notebook = wx.Notebook(hpanel, id=wx.ID_ANY, style=wx.BK_DEFAULT)
        hsizer.Add(notebook, 1, wx.ALL|wx.EXPAND, 5)
        self.preview = PreviewPanel(hpanel)
        hsizer.Add(self.preview, 1, wx.EXPAND)
        hpanel.SetSizer(hsizer)
        main_vsizer.Add(hpanel, 2, wx.EXPAND)

        self.run_panel = wx.Panel(main_panel)
        runsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.run_label = wx.StaticText(self.run_panel, label="Unchecked")
        self.run_label.SetFont(wx.Font(12, wx.DEFAULT, wx.NORMAL, wx.BOLD))
        runsizer.Add(self.run_label, 1, wx.EXPAND)
        self.run_btn = wx.Button(self.run_panel, label="Run")
        runsizer.Add(self.run_btn, 0, wx.ALIGN_CENTER_VERTICAL)
        self.run_panel.SetSizer(runsizer)
        main_vsizer.Add(self.run_panel, 0, wx.EXPAND)
        self.run_panel.Layout()

        main_panel.SetSizer(main_vsizer)

        self.run = AslRun(self, self.run_btn, self.run_label)
        setattr(self.run, "preview", self.preview)
        tab_cls = [AslInputOptions, StructureTab, AslCalibration, AslDistCorr, AslAnalysis,]
        tabs = [cls(notebook, idx, len(tab_cls)) for idx, cls in enumerate(tab_cls)]

        for tab in tabs:
            notebook.AddPage(tab, tab.title)
            setattr(tab, "run", self.run)
            setattr(tab, "preview", self.preview)
            setattr(self.run, tab.name, tab)
            setattr(self.preview, tab.name, tab)
            for tab2 in tabs:
                if tab != tab2:
                    setattr(tab, tab2.name, tab2)
            tab.update()

        self.Layout()

def main():
    """
    Program entry point
    """
    app = wx.App(redirect=False)
    top = AslGui()
    top.Show()
    app.MainLoop()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
BASIL - Bayesian model fitting for ASL

The BASIL module is a little more complex than the other Workspace based
modules because of the number of options available and the need for flexibility
in how the modelling steps are run.

The main function is ``basil`` which performs model fitting on ASL data 
in the Workspace ``asldata`` attribute. 

    wsp = Workspace()
    wsp.asldata = AslImage("asldata.nii.gz", tis=[1.6,])
    wsp.infertiss = True
    basil(wsp, output_wsp=wsp.sub("basil"))
    basil.finalstep.mean_ftiss.save("mean_ftiss.nii.gz")

Because of the number of options possible for the modelling process, the 
workspace attribute ``basil_options`` can be set as a dictionary of extra
options relevant only to Basil:

    wsp = Workspace()
    wsp.asldata = AslImage("asldata.nii.gz", tis=[1.6,])
    wsp.basil_options = {"infertiss" : True, "spatial" : True}
    basil(wsp, output_wsp=wsp.sub("basil"))
    basil.finalstep.mean_ftiss.save("mean_ftiss.nii.gz")

All options specified in basil_options are either consumed by Basil, or
if not passed directly to the model.

Copyright (c) 2008-2018 University of Oxford
"""

import sys
import math

import numpy as np

from fsl.wrappers import LOAD
from fsl.data.image import Image

from oxasl import __version__, __timestamp__, Workspace, image
from oxasl.options import AslOptionParser, OptionCategory, IgnorableOptionGroup, GenericOptions

def basil(wsp, prefit=True):
    wsp.log.write("\nRunning BASIL Bayesian modelling on ASL data\n")

    if wsp.asldata is None:
        raise ValueError("Input ASL data is None")

    wsp.log.write("  - BASIL v%s\n" % __version__)
    wsp.asldata.summary(log=wsp.log)

    if prefit and max(wsp.asldata.rpts) > 1:
        # Initial BASIL run on mean data
        wsp.log.write(" - Doing initial fit on mean at each TI\n\n")
        init_wsp = wsp.sub("init")
        init_wsp.asldata = wsp.asldata.mean_across_repeats()
        runbasil(init_wsp)
        main_wsp = wsp.sub("main")
        main_wsp.initmvn = wsp.init.finalstep.finalMVN
    else:
        main_wsp = wsp

    # Main run on full ASL data
    wsp.log.write("\n - Doing fit on full ASL data\n\n")
    runbasil(main_wsp)

    # Provide a reference to the final step
    wsp.finalstep = main_wsp.finalstep

def runbasil(wsp):
    """
    Run Bayesian model fitting on ASL data

    See ``basil`` for details of workspace attributes used

    :param wsp: Workspace object
    :param asldata: AslImage object to use as input data
    :param output_wsp: Optional Workspace object for storing output files. If not specified
                       ``wsp`` is used instead
    """
    steps = basil_steps(wsp)

    prev_result = None
    wsp.asldata_diff = wsp.asldata.diff().reorder("rt")

    for idx, step in enumerate(steps):
        step_wsp = wsp.sub("step%i" % (idx+1))
        desc = "Step %i of %i: %s" % (idx+1, len(steps), step.desc)
        if prev_result is not None:
            desc += " - Initialise with step %i" % idx
        step_wsp.log.write(desc + "     ")
        result = step.run(prev_result, log=wsp.log)
        for key, value in result.items():
            setattr(step_wsp, key, value)

        if step_wsp.logfile is not None and step_wsp.savedir is not None:
            step_wsp.set_item("logfile", step_wsp.logfile, save_fn=str)

        prev_result = result
    wsp.finalstep = step_wsp
    wsp.log.write("\nEnd\n")

def basil_steps(wsp):
    # Get data in the format expected by BASIL
    asldata = wsp.asldata.diff().reorder("rt")

    # Default Fabber options for VB runs and spatial steps. Note that attributes
    # which are None (e.g. sliceband) are not passed to Fabber
    options = {
        "data" : asldata,
        "mask" : wsp.rois.mask,
        "model" : "aslrest",
        "disp" : "none",
        "exch" : "mix",
        "method" : "vb",
        "noise" : "white",
        "allow-bad-voxels" : True,
        "max-iterations" : 20,
        "convergence" : "trialmode",
        "max-trials" : 10,
        "save-mean" : True,
        "save-mvn" : True,
        "save-std" : True,
        "save-model-fit" : True,
        "t1b" : 1.65,
    }

    # Default is to infer tissue parameters only
    infertiss = wsp.ifnone("infertiss", True)
    inferbat = wsp.ifnone("inferbat", False)
    inferart = wsp.ifnone("inferart", False)
    inferpc = wsp.ifnone("inferpc", False)
    infert1 = wsp.ifnone("infert1", False)

    # Default to inferring exch/disp parameters if a relevant model is used
    inferdisp = wsp.ifnone("inferdisp", options["disp"] != "none")
    inferexch = wsp.ifnone("inferexch", options["exch"] != "mix")

    # if we are doing CASL then fix the bolus duration, unless explicitly told us otherwise
    infertau = wsp.ifnone("infertau", not asldata.casl)
    
    # Single or Multi TI setup
    if asldata.ntis == 1:
        # Single TI data - don't try to infer arterial component of bolus duration, we don't have enough info
        wsp.log.write(" - Operating in Single TI mode - no arterial component, fixed bolus duration\n")
        inferart = False
        infertau = False
        options["batsd"] = 0.1
    else:
        # For multi TI/PLD data, set a more liberal prior for tissue ATT since we should be able to 
        # determine this from the data. NB this leaves the arterial BAT alone.
        options["batsd"] = 1

    if wsp.wp:
        # White paper mode - this overrides defaults, but can be overwritten by command line 
        # specification of individual parameters
        wsp.log.write(" - Analysis in white paper mode: T1 default=1.65, BAT default=0, voxelwise calibration\n")
        options["t1"] = 1.65
        options["bat"] = 0.0
    else:
        options["t1"] = 1.3
        if asldata.casl:
            options["bat"] = 1.3
        else:
            options["bat"] = 0.7

    # Override defaults using workspace arguments
    for attr in ("t1", "t1b", "FA", "bat", "batsd", "exch", "disp"):
        value = getattr(wsp, attr)
        if value is not None:
            options[attr] = value

    # We choose to pass TIs (not PLDs). The asldata object ensures that
    # TIs are correctly derived from PLDs, when these are specified, by adding
    # the bolus duration.
    for idx, ti in enumerate(asldata.tis):
        options["ti%i" % (idx+1)] = ti
        options["rpt%i" % (idx+1)] = asldata.rpts[idx]

    # Bolus duration - use a single value where possible as cannot infer otherwise
    taus = getattr(asldata, "taus", [1.8,])
    if min(taus) == max(taus):
        options["tau"] = taus[0]
    else:
        for idx, tau in enumerate(taus):
            options["tau%i" % (idx+1)] = tau

    # Other asl data parameters
    for attr in ("casl", "slicedt", "sliceband"):
        if getattr(asldata, attr, None) is not None:
            options[attr] = getattr(asldata, attr)

    if wsp.noiseprior:
        # Use an informative noise prior
        if wsp.noisesd is None:
            snr = wsp.ifnone("snr", 10)
            wsp.log.write(" - Using SNR of %f to set noise std dev\n" % snr)

            # Estimate signal magntiude FIXME diffdata_mean is always 3D?
            diffdata_mean = asldata.mean()
            if diffdata_mean.ndim > 3:
                datamax = np.amax(diffdata_mean.data, 3)
            else:
                datamax = diffdata_mean.data
            brain_mag = np.mean(datamax.data[wsp.rois.mask.data != 0])
            # this will correspond to whole brain CBF (roughly) - about 0.5 of GM
            noisesd = math.sqrt(brain_mag * 2 / snr)
        else:
            noisesd = wsp.noisesd
        wsp.log.write(" - Using a prior noise sd of: %f\n" % noisesd)
        options["prior-noise-stddev"] = noisesd

    # Options for final spatial step
    prior_type_spatial = "M"
    prior_type_mvs = "A"
    options_svb = {
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "convergence" : "maxits",
        "max-iterations": 20,
    }

    wsp.log.write("Model (in fabber) is : %s\n" % options["model"])
    wsp.log.write("Dispersion model option is %s\n" % options["disp"])
    wsp.log.write("Compartment exchange model option is %s\n" % options["exch"])

    # Partial volume correction
    pvcorr = wsp.pwm is not None or wsp.pgm is not None
    if pvcorr:
        if wsp.pwm is None or wsp.pgm is None:
            raise ValueError("Only one partial volume map (GM / WM) was supplied for PV correctioN")
        # Need a spatial step with more iterations for the PV correction
        options_svb["max-iterations"] = 200
        # Ignore partial volumes below 0.1
        pgm = np.copy(wsp.pgm.data)
        pwm = np.copy(wsp.pwm.data)
        pgm[pgm < 0.1] = 0
        pgm[pgm > 1] = 1
        pwm[pwm < 0.1] = 0
        pwm[pwm > 1] = 1
        pgm = Image(pgm, header=wsp.pgm.header)
        pwm = Image(pwm, header=wsp.pwm.header)
        
    if pvcorr and not infertiss:
        raise ValueError("ERROR: PV correction is not compatible with --artonly option (there is no tissue component)")

    # Set general parameter inference and inclusion
    if infertiss:
        options["inctiss"] = True
    if inferbat:
        options["incbat"] = True
        options["inferbat"] = True # Infer in first step
    if inferart:
        options["incart"] = True
    if inferpc:
        options["incpc"] = True
    if infertau:
        options["inctau"] = True
    if infert1:
        options["inct1"] = True
    if pvcorr:
        options["incpve"] = True

    # Keep track of the number of spatial priors specified by name
    spriors = 1 

    if wsp.initmvn:
        # we are being supplied with an initial MVN
        wsp.log.write("Initial MVN being loaded %s\n" % wsp.initmvn.name)
        options["continue-from-mvn"] = wsp.initmvn
    
    # T1 image prior
    if wsp.t1im:
        spriors = _add_prior(options, spriors, "T_1", type="I", image=wsp.t1im)

    steps = []
    components = ""

    ### --- TISSUE MODULE ---
    if infertiss:
        components += " Tissue "
        options["infertiss"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

        # setup spatial priors ready
        spriors = _add_prior(options_svb, spriors, "ftiss", type=prior_type_spatial)

    ### --- ARTERIAL MODULE ---
    if inferart:
        components += " Arterial "
        options["inferart"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

        # setup spatial priors ready
        spriors = _add_prior(options_svb, spriors, "fblood", type=prior_type_mvs)

    ### --- BOLUS DURATION MODULE ---
    if infertau:
        components += " Bolus duration "
        options["infertau"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

    ### --- MODEL EXTENSIONS MODULE ---
    # Add variable dispersion and/or exchange parameters and/or pre-capiliary
    if inferdisp or inferexch or wsp.inferpc:
        if inferdisp:
            components += " dispersion"
            options["inferdisp"] = True
        if inferexch:
            components += " exchange"
            options["inferexch"] = True
        if inferpc:
            components += " pre-capiliary"
            options["inferpc"] = True

        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

    ### --- T1 MODULE ---
    if infert1:
        components += " T1 "
        options["infert1"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

    ### --- PV CORRECTION MODULE ---
    if pvcorr:
        # setup ready for PV correction, which has to be done with spatial priors
        components += " PVE"
        options["pvcorr"] = True

        # set the image priors for the PV maps
        spriors = _add_prior(options, spriors, "pvgm", type="I", image=pgm)
        spriors = _add_prior(options, spriors, "pvwm", type="I", image=pwm)
        spriors = _add_prior(options, spriors, "fwm", type="M")

        if steps:
            # Add initialisaiton step for PV correction - ONLY if we have something to init from
            steps.append(PvcInitStep({"data" : asldata, "mask" : wsp.rois.mask, "pgm" : pgm, "pwm" : pwm}, "PVC initialisation"))

    ### --- SPATIAL MODULE ---
    if wsp.spatial:
        step_desc = "Spatial VB - %s" % components
        options.update(options_svb)
        del options["max-trials"]

        if not wsp.onestep:
            steps.append(FabberStep(options, step_desc))

    ### --- SINGLE-STEP OPTION ---
    if wsp.onestep:
        steps.append(FabberStep(options, step_desc))
        
    if not steps:
        raise ValueError("No steps were generated - no parameters were set to be inferred")
        
    return steps

def _add_prior(options, prior_idx, param, **kwargs):
    options["PSP_byname%i" % prior_idx] = param
    for key, value in kwargs.items():
        options["PSP_byname%i_%s" % (prior_idx, key)] = value
    return prior_idx + 1

class Step(object):
    """
    A step in the Basil modelling process
    """
    def __init__(self, options, desc):
        self.options = dict(options)
        self.desc = desc

class FabberStep(Step):
    """
    A Basil step which involves running Fabber
    """
    def run(self, prev_output, log=sys.stdout):
        """
        Run Fabber, initialising it from the output of a previous step
        """
        if prev_output is not None:
            self.options["continue-from-mvn"] = prev_output["finalMVN"]
        from .wrappers import fabber
        ret = fabber(self.options, output=LOAD, progress=log)
        log.write("\n")
        return ret

class PvcInitStep(Step):
    """
    A Basil step which initialises partial volume correction
    """
    def run(self, prev_output, log=sys.stdout):
        """
        Update the MVN from a previous step to include initial estimates
        for PVC parameters
        """
        log.write("Initialising partial volume correction...\n")
        mask = self.options["mask"]
        # set the inital GM amd WM values using a simple PV correction
        wm_cbf_ratio = 0.4

        # Modified pvgm map
        #fsl.maths(pgm, " -sub 0.2 -thr 0 -add 0.2 temp_pgm")
        temp_pgm = np.copy(self.options["pgm"].data)
        temp_pgm[temp_pgm < 0.2] = 0.2

        # First part of correction psuedo WM CBF term
        #fsl.run("mvntool", "--input=temp --output=temp_ftiss --mask=%s --param=ftiss --param-list=step%i/paramnames.txt --val" % (mask.name, prev_step))
        #fsl.maths("temp_ftiss", "-mul %f -mul %s wmcbfterm" % (wm_cbf_ratio, pwm.name))
        prev_ftiss = prev_output["mean_ftiss"].data
        wm_cbf_term = (prev_ftiss * wm_cbf_ratio) * self.options["pwm"].data

        #fsl.maths("temp_ftiss", "-sub wmcbfterm -div temp_pgm gmcbf_init")
        #fsl.maths("gmcbf_init -mul %f wmcbf_init" % wm_cbf_ratio)
        gmcbf_init = (prev_ftiss - wm_cbf_term) / temp_pgm
        wmcbf_init = gmcbf_init * wm_cbf_ratio

        mvn = prev_output["finalMVN"]
        gmcbf_init = Image(gmcbf_init, header=mvn.header)
        wmcbf_init = Image(wmcbf_init, header=mvn.header)

        # load these into the MVN
        mvn = prev_output["finalMVN"]
        from .wrappers import mvntool
        params = prev_output["paramnames"]
        mvn = mvntool(mvn, params.index("ftiss")+1, output=LOAD, mask=mask, write=True, valim=gmcbf_init, var=0.1)["output"]
        mvn = mvntool(mvn, params.index("fwm")+1, output=LOAD, mask=mask, write=True, valim=wmcbf_init, var=0.1)["output"]
        log.write("DONE\n")
        return {"finalMVN" : mvn, "gmcbf_init" : gmcbf_init, "wmcbf_init" : wmcbf_init}

class BasilOptions(OptionCategory):
    """
    BASIL option category
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "basil", **kwargs)

    def groups(self, parser):
        groups = []
        
        group = IgnorableOptionGroup(parser, "BASIL options", ignore=self.ignore)
        group.add_option("--infertau", help="Infer bolus duration", action="store_true", default=False)
        group.add_option("--inferart", help="Infer macro vascular (arterial) signal component", action="store_true", default=False)
        group.add_option("--inferpc", help="Infer pre-capillary signal component", action="store_true", default=False)
        group.add_option("--infert1", help="Include uncertainty in T1 values", action="store_true", default=False)
        group.add_option("--artonly", help="Remove tissue component and infer only arterial component", action="store_true", default=False)
        group.add_option("--fixbat", dest="inferbat", help="Fix bolus arrival time", action="store_false", default=True)
        group.add_option("--spatial", help="Add step that implements adaptive spatial smoothing on CBF", action="store_true", default=False)
        group.add_option("--fast", help="Faster analysis (1=faster, 2=single step", type=int, default=0)
        group.add_option("--noiseprior", help="Use an informative prior for the noise estimation", action="store_true", default=False)
        group.add_option("--noisesd", help="Set a custom noise std. dev. for the nosie prior", type=float)
        group.add_option("--model-options", help="File containing additional model-specific options")
        groups.append(group)

        group = IgnorableOptionGroup(parser, "Model options", ignore=self.ignore)
        group.add_option("--disp", help="Model for label dispersion", default="none")
        group.add_option("--exch", help="Model for tissue exchange (residue function)", default="mix")
        groups.append(group)

        group = IgnorableOptionGroup(parser, "Partial volume correction / CBF estimation (enforces --spatial)", ignore=self.ignore)
        group.add_option("--pgm", help="Gray matter PV map", type="image")
        group.add_option("--pwm", help="White matter PV map", type="image")
        groups.append(group)

        group = IgnorableOptionGroup(parser, "Special options", ignore=self.ignore)
        group.add_option("--t1im", help="Voxelwise T1 tissue estimates", type="image")
        groups.append(group)

        return groups

def main():
    """
    Entry point for BASIL command line application
    """
    try:
        parser = AslOptionParser(usage="basil -i <ASL input file> [options...]", version=__version__)
        parser.add_category(image.AslImageOptions())
        parser.add_category(BasilOptions())
        parser.add_category(GenericOptions())
        
        options, _ = parser.parse_args()
        if not options.output:
            options.output = "basil"

        if not options.asldata:
            sys.stderr.write("Input file not specified\n")
            parser.print_help()
            sys.exit(1)
        
        wsp = Workspace(savedir=options.output, separate_input=True, auto_asldata=True, **vars(options))
        
        # Deal with --artonly
        if wsp.artonly:
            wsp.infertiss = False
            wsp.inferart = True

        # Adjust number of iterations based on fast option
        if not wsp.fast:
            num_iter, num_trials, onestep = 20, 10, False
        elif wsp.fast == 1:
            num_iter, num_trials, onestep = 5, 2, False
        elif wsp.fast == 2:
            num_iter, num_trials, onestep = 10, 5, True
        else:
            raise ValueError("Not a valid option for fast: %s" % str(wsp.fast))

        wsp.max_iterations = num_iter
        wsp.max_trials = num_trials
        wsp.onestep = onestep

        # Read in additional model options from a file
        wsp.model_options = {}
        if wsp.model_options:
            with open(wsp.model_options) as optfile:
                for line in optfile:
                    keyval = line.strip().rstrip("\n").lstrip("--").split("=", 1)
                    key = keyval[0].strip().replace("-", "_")
                    if key != "":
                        if len(keyval) == 1:
                            wsp.model_options[key] = True
                        else:
                            wsp.model_options[key] = keyval[1].strip()

        # Run BASIL processing
        basil(wsp)
        
    except ValueError as exc:
        sys.stderr.write("\nERROR: " + str(exc) + "\n")
        sys.stderr.write("Use --help for usage information\n")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/bin/env python
"""
ASL_REG: Registration for ASL data

Michael Chappell, IBME QuBIc & FMRIB Image Analysis Groups

Copyright (c) 2008-2018 University of Oxford
"""
from __future__ import unicode_literals

import os
import sys
import math
import warnings
import traceback

import numpy as np

from fsl.data.image import Image
import fsl.wrappers as fsl

from oxasl import __version__, Workspace, struc, brain
from oxasl.options import AslOptionParser, GenericOptions, OptionCategory, IgnorableOptionGroup, load_matrix
from oxasl.wrappers import epi_reg
from oxasl.reporting import LightboxImage

def init(wsp):
    """
    Create registration sub-workspace if not already there
    """
    if wsp.reg is None:
        wsp.sub("reg")

def get_regfrom(wsp):
    """
    Set the 3D image to be used as the ASL registration target for structural->ASL registration

    Regfrom defines the 'native' space

    Optional workspace attributes
    -----------------------------

     - ``regfrom`` : User-supplied registration reference image
     - ``asldata`` : Raw ASL data
     - ``calib``   : Calibration image

    Updated workspace attributes
    ----------------------------

     - ``regfrom``    : Registration reference image in ASL space
    """
    init(wsp)
    if wsp.reg.regfrom is None:
        wsp.log.write("\nGetting the ASL image to use for registration)\n")
        if wsp.regfrom is not None:
            wsp.log.write(" - Registration reference image supplied by user\n")
            wsp.reg.regfrom = wsp.regfrom
        elif wsp.asldata.iaf in ("tc", "ct"):
            wsp.log.write(" - Registration reference is mean ASL signal (brain extracted)\n")
            wsp.reg.regfrom = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
        elif wsp.calib is not None and wsp.calib.sameSpace(wsp.asldata):
            wsp.log.write(" - Registration reference is calibration image (brain extracted)\n")
            wsp.reg.regfrom = brain.brain(wsp, wsp.calib, thresh=0.2)
        else:
            wsp.log.write(" - Registration reference is mean ASL image (brain extracted)\n")
            wsp.reg.regfrom = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)

def get_motion_params(mat):
    """
    Get motion parameters from a Flirt motion correction matrix

    This is done under the assumption that the matrix may contain
    rotation, translation and possibly minor scaling but no reflection,
    shear etc. So the output could be incorrect for some extreme
    correction matrices, but this probably indicates an error in the
    registration process. We wrap the whole thing in a try block so
    if anything goes horribly wrong it does not at least stop the
    pipeline running

    See http://en.wikipedia.org/wiki/Rotation_matrix for details
    of the rotation calculation.

    :return: magnitude of translation, angle and rotation axis
    """
    if tuple(mat.shape) != (4, 4):
        raise ValueError("Not a 4x4 Flirt matrix")

    try:
        # Extract scales - last one is the magnitude of the translation
        scales = np.linalg.norm(mat[:3, :], axis=0)

        # Normalise unit vectors by scaling before extracting rotation
        mat[:, 0] /= scales[0]
        mat[:, 1] /= scales[1]
        mat[:, 2] /= scales[2]

        # Rotation axis
        rot_axis = np.array([
            mat[2, 1] - mat[1, 2],
            mat[0, 2] - mat[2, 0],
            mat[1, 0] - mat[0, 1],
        ], dtype=np.float)

        # Rotation angle - note that we need to check the sign
        trace = np.trace(mat[:3, :3])
        costheta = (float(trace)-1) / 2
        sintheta = math.sqrt(1-costheta*costheta)
        theta = math.acos(costheta)
        test_element = rot_axis[1]*rot_axis[0]*(1-costheta) + rot_axis[2]*sintheta
        if np.abs(test_element - mat[1, 0]) > np.abs(test_element - mat[0, 1]):
            theta = -theta
        return scales[-1], math.degrees(theta), rot_axis
    except:
        warnings.warn("Error extracting motion parameters from transformation matrix - check registration/moco looks OK!")
        traceback.print_exc()
        return 1, 0, [0, 0, 1]

def reg_asl2calib(wsp):
    """
    Register calibration image to ASL space

    Note that this might already have been done as part of motion correction
    """
    init(wsp)
    if wsp.calib is not None and wsp.reg.asl2calib is None:
        get_regfrom(wsp)
        wsp.log.write("Registering calibration image to ASL image\n")
        _, wsp.reg.asl2calib = reg_flirt(wsp, wsp.reg.regfrom, wsp.calib)
        wsp.reg.calib2asl = np.linalg.inv(wsp.reg.asl2calib)

def reg_asl2struc(wsp, flirt=True, bbr=False, name="initial"):
    """
    Registration of ASL images to structural image

    :param flirt: If provided, sets whether to use FLIRT registration
    :param bbr: If provided, sets whether to use BBR registration

    Required workspace attributes
    -----------------------------

     - ``regfrom``            : Registration reference image in ASL space
     - ``struc``              : Structural image

    Updated workspace attributes
    ----------------------------

     - ``asl2struc``    : ASL->structural transformation matrix
     - ``struc2asl``    : Structural->ASL transformation matrix
     - ``regto``        : ``regfrom`` image transformed to structural space
    """
    init(wsp)
    struc.init(wsp)
    if wsp.structural.struc is not None:
        get_regfrom(wsp)
        wsp.log.write("\nRegistering ASL data to structural data\n")
        if flirt:
            wsp.reg.regto, wsp.reg.asl2struc = reg_flirt(wsp, wsp.reg.regfrom, wsp.structural.brain, wsp.reg.asl2struc)
        if bbr:
            wsp.reg.regto, wsp.reg.asl2struc = reg_bbr(wsp)

        wsp.reg.struc2asl = np.linalg.inv(wsp.reg.asl2struc)

        wsp.log.write(" - ASL->Structural transform\n")
        wsp.log.write(str(wsp.reg.asl2struc) + "\n")
        wsp.log.write(" - Structural->ASL transform\n")
        wsp.log.write(str(wsp.reg.struc2asl) + "\n")

        page = wsp.report.page("asl2struc_%s" % name)
        page.heading("%s ASL -> Structural registration" % name.title(), level=0)
        page.heading("Transformation parameters", level=1)
        motion_params = get_motion_params(wsp.reg.asl2struc)
        page.table([
            ["Translation magnitude", "%.3g mm" % motion_params[0]],
            ["Rotation magnitude", "%.3g \N{DEGREE SIGN}" % motion_params[1]],
        ])
        page.heading("ASL->Structural transformation matrix", level=1)
        page.matrix(wsp.reg.asl2struc)
        page.heading("Structural->ASL transformation matrix", level=1)
        page.matrix(wsp.reg.struc2asl)

        if wsp.structural.gm_seg is not None:
            gm_asl = struc2asl(wsp, wsp.structural.gm_seg, interp="nn")
            page.heading("GM mask aligned with ASL data", level=1)
            page.image("gm_reg_%s" % name, LightboxImage(gm_asl, bgimage=wsp.reg.regfrom))
            wm_asl = struc2asl(wsp, wsp.structural.wm_seg, interp="nn")
            page.heading("WM mask aligned with ASL data", level=1)
            page.image("wm_reg_%s" % name, LightboxImage(wm_asl, bgimage=wsp.reg.regfrom))

def reg_struc2std(wsp, fnirt=False):
    """
    Determine structural -> standard space registration

    Optional workspace attributes
    -----------------------------

     - ``structural.struc``   : Structural image
     - ``fslanat``            : Path to existing FSLANAT data

    Updated workspace attributes
    ----------------------------

     - ``reg.struc2std``    : Structural->MNI transformation matrix - either warp image or FLIRT matrix
     - ``reg.std2struc``    : MNI->structural transformation - either warp image or FLIRT matrix
    """
    init(wsp)

    if wsp.reg.std2struc is not None:
        return

    if wsp.fslanat:
        warp = os.path.join(wsp.fslanat, "T1_to_MNI_nonlin_coeff.nii.gz")
        mat = os.path.join(wsp.fslanat, "T1_to_MNI_lin.mat")
        if os.path.isfile(warp):
            wsp.log.write(" - Using structural->std nonlinear transformation from FSL_ANAT\n")
            wsp.reg.struc2std = Image(warp, loadData=False)
        elif os.path.isfile(mat):
            wsp.log.write(" - Using structural->std linear transformation from FSL_ANAT\n")
            wsp.reg.struc2std = load_matrix(mat)

    if wsp.reg.struc2std is None:
        struc.init(wsp)
        wsp.log.write(" - Registering structural image to standard space using FLIRT\n")
        flirt_result = fsl.flirt(wsp.structural.brain, os.path.join(os.environ["FSLDIR"], "data/standard/MNI152_T1_2mm_brain"), omat=fsl.LOAD)
        wsp.reg.struc2std = flirt_result["omat"]

        if fnirt:
            wsp.log.write(" - Registering structural image to standard space using FNIRT\n")
            fnirt_result = fsl.fnirt(wsp.structural.brain, aff=wsp.reg.struc2std, config="T1_2_MNI152_2mm.cnf", cout=fsl.LOAD)
            wsp.reg.struc2std = fnirt_result["cout"]

    if isinstance(wsp.reg.struc2std, Image):
        # Calculate the inverse warp using INVWARP
        invwarp_result = fsl.invwarp(wsp.reg.struc2std, wsp.structural.struc, out=fsl.LOAD)
        wsp.reg.std2struc = invwarp_result["out"]
    else:
        wsp.reg.std2struc = np.linalg.inv(wsp.reg.struc2std)

def std2struc(wsp, img, **kwargs):
    """
    Transform an image from standard space to structural space
    """
    return transform(wsp, img, wsp.reg.std2struc, wsp.structural.struc, **kwargs)

def struc2std(wsp, img, **kwargs):
    """
    Transform an image from structural space to standard space
    """
    ref = Image(os.path.join(os.environ["FSLDIR"], "data/standard/MNI152_T1_2mm_brain"))
    return transform(wsp, img, wsp.reg.struc2std, ref, **kwargs)

def struc2asl(wsp, img, **kwargs):
    """
    Convert an image from structural to ASL space

    :param img: Image object in structural space
    :return: Transformed Image object in ASL (native) space
    """
    init(wsp)
    return transform(wsp, img, wsp.reg.struc2asl, wsp.nativeref, **kwargs)

def asl2struc(wsp, img, **kwargs):
    """
    Convert an image from ASL to structural space

    Keyword arguments are passed to ``transform``

    :param img: Image object in native (ASL) space
    :return: Transformed Image object in structural space
    """
    init(wsp)
    return transform(wsp, img, wsp.reg.asl2struc, wsp.structural.struc, **kwargs)

def calib2asl(wsp, img, **kwargs):
    """
    Convert an image from calibration space to ASL space

    :param img: Image object in calibration space
    :return: Transformed Image object in ASL (native) space
    """
    init(wsp)
    return transform(wsp, img, wsp.reg.calib2asl, wsp.nativeref, **kwargs)

def asl2calib(wsp, img, **kwargs):
    """
    Convert an image from ASL to calibration space

    Keyword arguments are passed to ``transform``

    :param img: Image object in native (ASL) space
    :return: Transformed Image object in calibration space
    """
    init(wsp)
    return transform(wsp, img, wsp.reg.asl2calib, wsp.structural.struc, **kwargs)

def transform(wsp, img, trans, ref, use_flirt=False, interp="trilinear", paddingsize=1, premat=None, mask=False, mask_thresh=0.5):
    """
    Transform an image

    :param wsp: Workspace, used for logging only
    :param img: Image to transform
    :param trans: Transformation matrix or warp image
    :param ref: Reference image
    :param use_flirt: Use flirt to apply the transformation which must be a matrix
    :param interp: Interpolation method
    :param paddingsize: Padding size in pixels
    :param premat: If trans is a warp, this can be set to a pre-warp affine transformation matrix

    :return: Transformed Image object
    """
    if trans is None:
        raise ValueError("Transformation matrix not available - has registration been performed?")

    have_warp = isinstance(trans, Image)
    if use_flirt and have_warp:
        raise ValueError("Cannot transform using Flirt when we have a warp")
    elif use_flirt:
        if interp == "nn":
            interp = "nearestneighbour"
        ret = fsl.applyxfm(img, ref, trans, out=fsl.LOAD, interp=interp, paddingsize=paddingsize, log=wsp.fsllog)["out"]
    else:
        if have_warp:
            kwargs = {"warp" : trans, "premat" : premat, "rel" : True}
        elif premat:
            raise ValueError("Can't set a pre-transformation matrix unless using a warp")
        else:
            kwargs = {"premat" : trans}
        ret = fsl.applywarp(img, ref, out=fsl.LOAD, interp=interp, paddingsize=paddingsize, super=True, superlevel="a", log=wsp.fsllog, **kwargs)["out"]
    if mask:
        # Binarise mask images
        ret = Image((ret.data > mask_thresh).astype(np.int), header=ret.header)
    return ret

def reg_flirt(wsp, img, ref, initial_transform=None):
    """
    Register low resolution ASL or calibration data to a high resolution
    structural image using Flirt rigid-body registration

    The brain extracted structural image is used as the reference image. If
    this is not supplied, BET will be run on the whole head structural image.

    :param reg_img: Data to register, e.g. PWI or calibration image. Normally would be brain extracted
    :param struc_brain_img: Brain-extracted structural image

    Optional keyword arguments:

    :param inweight:
    :param init: Initial transform matrix
    :param schedule: FLIRT transform schedule file (default: xyztrans.sch")
    :param dof: FLIRT degrees of freedom

    :return Tuple of registered image, transform matrix
    """
    wsp.log.write(" - Registering image: %s using FLIRT\n" % img.name)

    # Step 1: 3D translation only
    flirt_opts = {
        "schedule" : os.path.join(os.environ["FSLDIR"], "etc", "flirtsch", "xyztrans.sch"),
        "init" : initial_transform,
        "inweight" : wsp.inweight,
        "log" : wsp.fsllog,
    }
    step1_trans = fsl.flirt(img, ref, omat=fsl.LOAD, **flirt_opts)["omat"]

    # Step 2: 6 DOF transformation with small search region
    flirt_opts.update({
        "schedule" : os.path.join(os.environ["FSLDIR"], "etc", "flirtsch", wsp.ifnone("flirtsch", "simple3D.sch")),
        "init" : step1_trans,
        "dof" : wsp.ifnone("dof", 6),
    })
    flirt_result = fsl.flirt(img, ref, out=fsl.LOAD, omat=fsl.LOAD, **flirt_opts)

    return flirt_result["out"], flirt_result["omat"]

def reg_bbr(wsp):
    """
    Perform BBR registration

    :param reg_img: Data to register, e.g. PWI or calibration image. Normally would be brain extracted
    :param struc_img: Structural image
    :param struc_brain_img: Brain-extracted structural image

    Optional keyword arguments:

    :param inweight:
    :param init: Initial transform matrix

    Optional keyword arguments for fieldmap distortion correction:

    :param fmap: Fieldmap image
    :param fmapmag: Fieldmap magnitude image
    :param fmapmagbrain: Fieldmap magnitude image - brain extracted
    :param pedir: Phase encoding direction (x, -x, y, -y, z, -z)
    :param echospacing: Echo spacing

    :return Tuple of registered image, transform matrix
    """
    struc.segment(wsp)

    wsp.log.write("  - BBR registration using epi_reg\n")
    # Windows can't run epi_reg as it's a batch script. Use our experimental python
    # implementation but use the standard epi_reg on other platforms until the python
    # version is better tested
    if sys.platform.startswith("win"):
        import oxasl.epi_reg as pyepi
        result = pyepi.epi_reg(wsp, wsp.reg.regfrom)
    else:
        result = epi_reg(epi=wsp.reg.regfrom, t1=wsp.structural.struc, t1brain=wsp.structural.brain, out=fsl.LOAD, wmseg=wsp.structural.wm_seg, init=wsp.reg.asl2struc, inweight=wsp.inweight, log=wsp.fsllog)
    return result["out.nii.gz"], result["out"]

    #OUTPUT
    #echo "Saving FINAL output"
    #if [ -z $finalonly ]; then
	#cp $outdir/asl2struct.mat $outdir/asl2struct_init.mat # save the initial transformation matrix to allow chekcing if this part failed
    #fi
    #cp $tempdir/low2high_final.mat $outdir/asl2struct.mat #the transformation matrix from epi_reg - this overwrites the version from MAIN registration
    #convert_xfm -omat $outdir/struct2asl.mat -inverse $outdir/asl2struct.mat #often useful to have the inverse transform, so calcuate it
    #if [ ! -z $fmap ]; then
	#imcp $tempdir/low2high_final_warp $outdir/asl2struct_warp #the warp from epi_reg
    #fi
    #imcp $tempdir/low2high_final $outdir/asl2struct # save the transformed image to check on the registration
    #
    # # copy the edge image from epi_reg output as that is good for visualisation
    # imcp $wm_seg $outdir/wm_seg
    #imcp $tempdir/low2high_final_fast_wmedge $outdir/tissedge

class RegOptions(OptionCategory):
    """
    OptionCategory which contains options for registration of ASL data to structural image
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "reg", **kwargs)

    def groups(self, parser):
        groups = []

        group = IgnorableOptionGroup(parser, "Registration", ignore=self.ignore)
        group.add_option("--regfrom", help="Registration image (e.g. perfusion weighted image)", type="image")
        #group.add_option("--omat", help="Output file for transform matrix", default=None)
        #group.add_option("--bbr", dest="do_bbr", help="Include BBR registration step using EPI_REG", action="store_true", default=False)
        #group.add_option("--flirt", dest="do_flirt", help="Include rigid-body registration step using FLIRT", action="store_true", default=True)
        #group.add_option("--flirtsch", help="user-specified FLIRT schedule for registration")
        groups.append(group)

        #group = IgnorableOptionGroup(parser, "Extra BBR registration refinement", ignore=self.ignore)
        #group.add_option("-c", dest="cfile", help="ASL control/calibration image for initial registration - brain extracted")
        #group.add_option("--wm_seg", dest="wm_seg", help="tissue segmenation image for bbr (in structural image space)")
        #groups.append(group)

        #group = IgnorableOptionGroup(parser, "Distortion correction using fieldmap (see epi_reg)", ignore=self.ignore)
        #g.add_option("--nofmapreg", dest="nofmapreg", help="do not perform registration of fmap to T1 (use if fmap already registered)", action="store_true", default=False)
        #groups.append(group)

        #group = IgnorableOptionGroup(parser, "Deprecated", ignore=self.ignore)
        #g.add_option("-r", dest="lowstruc", help="extra low resolution structural image - brain extracted")
        #g.add_option("--inweight", dest="inweight", help="specify weights for input image - same functionality as the flirt -inweight option", type="float")
        #groups.append(group)

        return groups

def main():
    """
    Entry point for command line tool
    """
    try:
        parser = AslOptionParser(usage="asl_reg [options]", version=__version__)
        parser.add_category(RegOptions())
        parser.add_category(struc.StructuralImageOptions())
        parser.add_category(GenericOptions())

        options, _ = parser.parse_args(sys.argv)
        wsp = Workspace(**vars(options))

        if not options.regfrom:
            sys.stderr.write("Input file not specified\n")
            parser.print_help()
            sys.exit(1)

        reg_asl2struc(wsp, wsp.do_flirt, wsp.do_bbr)
        if wsp.output:
            wsp.reg.regto.save(wsp.output)
        if wsp.reg.asl2struc:
            with open(wsp.omat, "w") as transform_file:
                for row in wsp.reg.asl2struc:
                    transform_file.write(" ".join(["%f" % val for val in row]) + "\n")

    except ValueError as exc:
        sys.stderr.write("ERROR: " + str(exc) + "\n")
        sys.exit(1)

if __name__ == "__main__":
    main()
